#! /usr/bin/env python
# -*- coding: utf-8 -*-

import tempfile

#################################
#
#   MAIN 
#
#################################
if  __name__ == "__main__":
    import subprocess
    import argparse
    import pysedm
    import numpy as np
    # ================= #
    #   Options         #
    # ================= #
    parser = argparse.ArgumentParser(
        description="""tool to subprocess wavelength solution build.
            """, formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('infile', type=str, default=None,
                        help='The date YYYYMMDD')

    # --------------- #
    #  Wavesoltuion   #
    # --------------- #
    parser.add_argument('--nsub',  type=str, default=2,
                        help='Number of subprocess to launch. [default 2]')
    
    parser.add_argument('--merge',  action="store_true", default=False,
                        help='Set this keyword to perform the Wavesolution merging. The rest will be ignored.')
    # --------------- #
    #  Subprocessing  #
    # --------------- #
    parser.add_argument('--spaxelrange', type=str, default="None",
                        help='Provide a range of spaxel indexe A,B ; only traces with index i>=A and i<B will be loaded. Indicated in saved filename.')
    
    parser.add_argument('--wavesolplots', action="store_true", default=False,
                        help='Set this to save individual wavelength solution fit results')

    parser.add_argument('--nofig', action="store_true", default=False,
                        help='')
    
    parser.add_argument('--rebuild',  action="store_true", default=False,
                        help='If the object you want to build already exists, nothing happens except if this is set')
    
    # - End
    args = parser.parse_args()
    
    # ================= #
    #   The Scripts     #
    # ================= #

    # --------- #
    #  Date     #
    # --------- #
    date = args.infile

    if args.merge:
        from pysedm.wavesolution import merge_wavesolutions
        wsol = merge_wavesolutions( pysedm.load_nightly_wavesolution(date,True))
        wsol.writeto( pysedm.io.get_datapath(date)+"%s_WaveSolution.pkl"%date)
        if not args.nofig:
            hgrid = pysedm.load_nightly_hexagonalgrid(date)
            wsol.show_dispersion_map(hgrid, vmax="98", vmin="2", outlier_highlight=5,
                            savefile= pysedm.io.get_datapath(date)+"%s_wavesolution_dispersionmap.pdf"%date)
            wsol.show_dispersion_map(hgrid, vmax="98", vmin="2", outlier_highlight=5,
                            savefile= pysedm.io.get_datapath(date)+"%s_wavesolution_dispersionmap.png"%date)

        wsol.calc_dispersion_stats(savefile=pysedm.io.get_datapath(date) +
                                   "%s_wavesolution_stats.txt" % date)
        import sys
        sys.exit(0)

    nsplit = int(args.nsub)
    
    # - TraceMatch Indexes
    tmap = pysedm.load_nightly_tracematch(date, withmask=False)
    idxall = tmap.get_traces_within_polygon( pysedm.sedm.INDEX_CCD_CONTOURS)
    
    # Split it in equal parts:
    k, m      = divmod( np.nanmax(idxall)+10, nsplit)
    idxbounds = [[i * k + min(i, m),(i + 1) * k + min(i + 1, m)] for i in range(nsplit)]
    
    
    options = []
    if args.nofig:
        options += ["--nofig"]
    if args.wavesolplots:
        options += ["--wavesolplots"]
    if args.spaxelrange not in ['None']:
        print(args.spaxelrange)
        options += ["--spaxelrange","%s,%s"%(args.spaxelrange[0],args.spaxelrange[1])]
    if args.rebuild:
        print(args.rebuild)
        options += ["--rebuild"]

   
    processes = []     
    for i,bounds in enumerate(idxbounds):          
        command = ["ccd_to_cube.py", date, "--wavesol","--spaxelrange","%s,%s"%(bounds[0],bounds[1])] + options
        print("launching command: "+" ".join(command))
        fd, path = tempfile.mkstemp()
        f = open(path, "w")
        p = subprocess.Popen(command, stdout=f)
        processes.append((p,f))

    for (p, f) in processes:
        p.wait()
        f.seek(0)
        # TODO log f here
        f.close()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from glob import glob
#################################
#
#   MAIN 
#
#################################
if  __name__ == "__main__":
    
    import argparse
    from pysedm.script.ccd_to_cube import *
    # ================= #
    #   Options         #
    # ================= #
    parser = argparse.ArgumentParser(
        description="""pysedm pipeline to build the cubebuilder objects
            """, formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('infile', type=str, default=None,
                        help='The date YYYYMMDD')

    parser.add_argument('--rebuild',  action="store_true", default=False,
                        help='If the object you want to build already exists, nothing happens except if this is set')
    
    parser.add_argument('--noguider',  action="store_true", default=False,
                        help='Avoid having a guider stack created. [part of the --build]')
    
    parser.add_argument('--solvewcs',  action="store_true", default=False,
                        help='Shall the wcs solution of the guider be solved (ignored if --noguider). [part of the --build]')

    parser.add_argument('--ncore',  type=int, default=None,
                        help='Number of cores to use for multiprocessing. ncore=1 means no multiprocessing.')

    # --------------- #
    #  Cube Building  #
    # --------------- #
    parser.add_argument('--build',  type=str, default=None,
                        help='Build a e3d cube of the given target (accepting regex) or target list (csv) e.g. --build dome or --build dome,Hg,Cd')

    parser.add_argument('--quickbuild', action="store_true", default=False,
                        help='Faster cube extraction: No background No j-flexure, no i-flexure. Keyword "quickbuild" added to the cube.')

    parser.add_argument('--nobackground', action="store_true", default=False,
                        help='Shall the ccd background be 0 instead of the usual pipeline background?')
    
    parser.add_argument('--noflexure', action="store_true", default=False,
                        help='build cubes without flexure correction')
    
    parser.add_argument('--notraceflexure', action="store_true", default=False,
                        help='build cubes without trace flexure (j-ccd correction)')
    
    parser.add_argument('--buildbkgd',  type=str, default=None,
                        help='Build a ccd background of the given target or target list (csv) e.g. --build dome or --build dome,Hg,Cd')

    parser.add_argument('--buildcal',  type=str, default=None,
                        help='Build the flux calibrated e3d cubes. Set this to "*" to use the --build arguments.')

    parser.add_argument('--calsource',  type=str, default=None,
                        help='The Inverse sensitivity spectrum used to calibrated the cubes. By default this uses the latest one.')
    
    # - Trace Matching
    parser.add_argument('--tracematch', action="store_true", default=False,
                        help='build the tracematch solution for the given night. This option saves masks (see tracematchnomasks)')
    
    parser.add_argument('--tracematchnomasks', action="store_true", default=False,
                        help='build te tracematch solution for the given night without saved the masks')
    
    # - Hexagonal Grid
    parser.add_argument('--hexagrid', action="store_true", default=False,
                        help='build the hexagonal grid (index<->qr<->xy) for the given night')

    # - Wavelength Solution
    parser.add_argument('--wavesol', action="store_true", default=False,
                        help='build the wavelength solution for the given night.')
    
    parser.add_argument('--spaxelrange', type=str, default="None",
                        help='Provide a range of spaxel indexe A,B ; only traces with index i>=A and i<B will be loaded. Indicated in saved filename.')
    
    parser.add_argument('--wavesoltest', type=str, default="None",
                        help='to be used with --wavesol. By setting --wavesoltest N one N random wavelength solution will be performed.')

    parser.add_argument('--wavesolplots', action="store_true", default=False,
                        help='Set this to save individual wavelength solution fit results')
    
    # ----------------- #
    #  Raw Calibration  #
    # ----------------- #
    parser.add_argument('--flat',    action="store_true", default=False,
                        help='Build the flat fielding for the night [see flatref to the reference object]')

    parser.add_argument('--flatref',  type=str, default="dome",
                        help='Build the flat fielding for the night ')
    
    parser.add_argument('--flatlbda',  type=str, default="5000,8500",
                        help='The wavelength range for the flat field. Format: min,max [in Angstrom] ')

    # ----------------- #
    #  Short Cuts       #
    # ----------------- #
    parser.add_argument('--allcalibs', action="store_true", default=False,
                        help='')
    
    parser.add_argument('--allscience', action="store_true", default=False,
                        help='')

    parser.add_argument('--nofig',    action="store_true", default=False,
                        help='')

    args = parser.parse_args()

    # Matplotlib
    # ================= #
    #   The Scripts     #
    # ================= #
    
    # --------- #
    #  Date     #
    # --------- #
    date = args.infile

    # ------------ #
    # Short Cuts   #
    # ------------ #
    if args.allcalibs:
        args.tracematch = True
        args.hexagrid   = True
        args.wavesol    = True
        args.build      = "dome"
        args.flat       = True

        
    # ================= #
    #   Actions         #
    # ================= #
            
    # - Builds
    if args.quickbuild:
        args.noflexure      = True
        args.notraceflexure = True
        args.nobackground   = True
        fileindex = "quickbuild"
    else:
        fileindex = ""
        
    if args.build is not None and len(args.build) >0:
        for target in args.build.split(","):
            build_night_cubes(date, target=target,
                             lamps=True, only_lamps=False, skip_calib=True,
                             fileindex=fileindex,
                             nobackground=bool(args.nobackground),
                             # - options
                             build_guider = False if args.noguider else True,
                             solve_wcs = args.solvewcs,
                             savefig = False if args.nofig else True,
                             flexure_corrected = False if args.noflexure else True,
                             traceflexure_corrected = False if args.notraceflexure else True,
                             ncore=args.ncore)
            
    if args.buildcal is not None:
        if args.buildcal=="*": args.buildcal=args.build
        if len(args.buildcal) >0:
            for target in args.buildcal.split(","):
                calibrate_night_cubes(date, target=target, 
                                    lamps=True, only_lamps=False, skip_calib=True)
                    
            
    # - Background
    if args.buildbkgd is not None and len(args.buildbkgd) > 0:
        for target in args.buildbkgd.split(","):
            build_backgrounds(date, target=target,
                                  show_progress=True,
                                  multiprocess=False, # Force no multiprocessing here
                                  lamps=True, only_lamps=True, skip_calib=True, 
                                  ncore=args.ncore)
            
    # -----------
    # 
    # ----------- 
    # - TraceMatch
    if args.tracematch or args.tracematchnomasks:
        build_tracematcher(date, save_masks= True if not args.tracematchnomasks else False,
                           rebuild=args.rebuild, ncore=args.ncore)
        
    # - Hexagonal Grid        
    if args.hexagrid:
        build_hexagonalgrid(date)
        
    # - Wavelength Solution
    if args.wavesol:
        ntest = None if "None" in args.wavesoltest else int(args.wavesoltest)
        spaxelrange = None if "None" in args.spaxelrange else np.asarray(args.spaxelrange.split(","), dtype="int")

        build_wavesolution(date, ntest=ntest, use_fine_tuned_traces=False,
                            idxrange=spaxelrange,
                            lamps=["Hg","Cd","Xe"], saveindividuals=args.wavesolplots,
                            savefig = False if args.nofig else True,
                            rebuild=args.rebuild)

    # - Flat Fielding
    if args.flat:
        lbda_min,lbda_max = np.asarray(args.flatlbda.split(","), dtype="float")
        build_flatfield(date,
                        lbda_min=lbda_min,
                        lbda_max=lbda_max,
                        ref=args.flatref, build_ref=True,
                        savefig=~args.nofig, ncore=args.ncore)
        # Now calc stats
        from pysedm import ccd
        from pysedm.io import get_datapath
        import numpy as np
        dome = ccd.get_dome("dome.fits", background=0, load_sep=True)
        a, b = dome.sepobjects.get(["a", "b"]).T
        savefile = get_datapath(date) + "%s_dome_stats.txt" % date
        stat_f = open(savefile, "w")
        stat_f.write("NSpax: %d\n" % len(b))
        stat_f.write("MinWid: %.3f\n" % min(b))
        stat_f.write("MaxWid: %.3f\n" % max(b))
        stat_f.write("MedWid: %.3f\n" % np.nanmedian(b))
        stat_f.write("AvgWid: %.3f\n" % np.nanmean(b))
        stat_f.write("MinLen: %.3f\n" % min(a))
        stat_f.write("MaxLen: %.3f\n" % max(a))
        stat_f.write("MedLen: %.3f\n" % np.nanmedian(a))
        stat_f.write("AvgLen: %.3f\n" % np.nanmean(a))
        stat_f.close()
        print("nspax, min, avg, med, max Wid: %d, %.3f, %.3f, %.3f, %.3f" %
              (len(b), min(b), np.nanmean(b), np.nanmedian(b), max(b)))



#! /usr/bin/env python
#

""" Module to pull pysedm data """


try:
    import ztfquery
    from ztfquery import sedm
except ImportError:
    raise ImportError("Please install ztfquery (v>1.6.0). pip install ztfquery")

# ================= #
#    Main object    #
# ================= #
SEDMQ = sedm.SEDMQuery()


# ======================== #
#                          #
#         MAIN             #
#                          #
# ======================== #
if  __name__ == "__main__":
    import argparse
    import numpy as np
    
    parser = argparse.ArgumentParser(
             description=""" pull SEDM data from pharos to your computer""",
             formatter_class=argparse.RawTextHelpFormatter)

    # - Infile - #
    parser.add_argument('infile', type=str, default=None,
                        help='Target Name')

    # ---------- #
    #  Options   #    
    # ---------- #
    parser.add_argument('--timerange',  type=str, default="2018-08-01,now",
                        help='You only one sedm data taken during that time range.'+"\n"+\
                             'format: coma separated values "a,b" \n Could be: '+"\n"+\
                             '- 1 time means starting point (here after the 12th of Sept 2018):  --timerange 2018-09-12  '+"\n"+\
                             '- 2 times means boundaries  (here between 12th and 15th of Sept 2018):  --timerange 2018-09-12,2018-09-15'+"\n"+\
                             '- for only upper limit  (here before 15th of Sept 2018):  --timerange None,2018-09-15'
                            )

    parser.add_argument('--info', action="store_true", default=False,
                        help='Get basic information concerning the given tartget'
                            )

        
    parser.add_argument('--data',  type=str, default="cube,spec",
                        help='Which kind of data do you want?'+"\n"+\
                             'format: coma separated values "cube,spec,crr"'+"\n"+
                             "NB: cube=e3d files ; spec = spectrum files ; ccd = ccd files "
                            )

    parser.add_argument('--extension',  type=str, default="*",
                        help='Which extension should be file have?'+"\n"+\
                             'format: coma separated values "pdf,fits,png"'
                            )
    parser.add_argument('--source',  type=str, default="local",
                        help='Where are you looking for the files?'+"\n"+\
                             'local: your computer | pharos: online"'
                            )
    
    parser.add_argument('--fileout',  type=str, default=None,
                        help='provide here a path to a file where you want to save the output.'+"\n"+\
                             'if None, the output will be printed'
                            )
    
    parser.add_argument('--quiet', action="store_true", default=False,
                        help='This prints information. Should this be quite?'
                            )
    
    args   = parser.parse_args()
    
    # ---------- #
    #  Target    #    
    # ---------- #
    target = args.infile
    if not args.quiet:
        print(" pysedm find ".center(40,'='))
        print("looking for data associated to: %s"%target)
        print("credit: based on ztfquery")
        
    # ---------- #
    #  Script    #    
    # ---------- #
    if args.info:
        import sys
        print(SEDMQ.get_target_data(target))
        sys.exit()
        
    
    # Downloading options
    datapath = SEDMQ.get_data_path(target, source=args.source,
                                    which=args.data.split(","),
                                    extension=args.extension if not "," in args.extension else args.extension.split(","),
                                    timerange=args.timerange.split(",") if args.timerange is not None else [None,None])
    if args.fileout is None or args.fileout in ["None"]:
        print(" results ".center(40,'='))
        print("\n".join(datapath))
    else:
        with open(args.fileout, 'w') as f:
            for data in datapath:
                f.write(data+"\n")
        
    



#! /usr/bin/env python
# -*- coding: utf-8 -*-

""" Module to I/O the data """

import os
import re
import warnings
from glob import glob
from astropy.io.fits import getheader, getval
from astropy.time import Time

try:
    from ztfquery.sedm import SEDMLOCAL_BASESOURCE
    REDUXPATH = os.path.join(SEDMLOCAL_BASESOURCE, "redux")
except:
    REDUXPATH   = os.getenv('SEDMREDUXPATH',default="~/redux/")

if REDUXPATH == "~/redux/":
    warnings.warn(f"Default REDUXPATH {REDUXPATH} is used.")
    
_PACKAGE_ROOT = os.path.abspath(os.path.dirname(__file__))+"/"
SEDM_REDUCER     = os.getenv('SEDM_USER',default="auto")

############################
#                          #
#  PROD/DB STRUCTURE       #
#                          #
############################
PROD_CUBEROOT        = "e3d"
PROD_SPECROOT        = "spec"
PROD_SENSITIVITYROOT = "fluxcal"

PRODSTRUCT_RE = {"ccd":{"lamp":"^(dome|Hg|Cd|Xe)",
                      "crr":'^(crr)',
                      "orig":"^(ifu)"},
               "bkgd":{"lamp":"^(bkgd).*(dome|Hg|Cd|Xe)",
                       "crr":"^(bkgd_crr)"},
               "cube": {"basic":"^(%s).((?!cal))"%PROD_CUBEROOT, # starts w/ e3d & not containing cal
                        "calibrated":"^(%s_cal)"%PROD_CUBEROOT,
                        "defaultcalibrated":"^(%s_defcal)"%PROD_CUBEROOT},
                "spec": {"basic":"^(%s).((?!cal))"%PROD_SPECROOT, # starts w/ e3d & not containing cal
                         "bkgd":"^(%s).((?!cal))"%PROD_SPECROOT, # starts w/ e3d & not containing cal
                         "auto":"^(%sauto).((?!cal))"%PROD_SPECROOT, # starts w/ e3d & not containing cal
                         "forcepsf":"^(%s_forcepsf).((?!cal))"%PROD_SPECROOT, # starts w/ e3d & not containing cal
                         "calibrated":"^(%s_cal)"%PROD_SPECROOT,
                         "defaultcalibrated":"^(%s_defcal)"%PROD_SPECROOT,
                         "invsensitivity":"^(%s)"%PROD_SENSITIVITYROOT},
                "psf": {"param":"^(%s)(?=.*json)"%"psf"} #.json files containing the psf fitvalues and fitted adr.
              }

__all__ = ["get_night_files",
               "load_nightly_mapper",
               "load_nightly_tracematch","load_nightly_hexagonalgrid",
               "load_nightly_wavesolution","load_nightly_flat"]

############################
#                          #
#  Data Access             #
#                          #
############################
def get_night_files(date, kind, target=None, extention=".fits"):
    """ GENERIC IO FUNCTION
    
    Parameters
    ----------
    date: [string] 
        date for which you want file. Format: YYYYMMDD
    
    kind: [string]
        Which kind of file do you want. 
        Format for kind:
         - You can directly provide a regular expression by starting by re:
            - e.g. all the lamp or dome associated data: kind="re:(dome|Hg|Xe|Cd)"
         - Ortherwise, you give provide a predefined format using 'type.subtype' ;
            - e.g. the basic cubes            : kind='cube.basic'
            - e.g. final flux calibrated cubes: kind='cube.fluxcalibrated' ;
            - e.g. quick flux calibrated cubes: kind='cube.defaultcalibrated' ;
            - e.g. the lamp or dome ccd files : kind='ccd.lamp'
            - e.g. cosmic ray corrected ccds  : kind='ccd.crr' ;
            - e.g. lamp and dome ccd files    : kind='ccd.lamp';

        (see all the predifined format in pysedm.io.PRODSTRUCT_RE) 
       
    target: [string/None] -optional-
        Additional selection. The file should also contain the string defined in target.
        target supports regex. e.g. requesting dome or Hg files: target="(dome|Hg)"

    Returns
    -------
    list of string (FullPaths
    """
    if kind.startswith('re:'):
        regex = kind.replace("re:","")
    else:
        if "." not in kind: kind = "%s.*"%kind
        
        kinds = kind.split(".")
        if len(kinds)==2:
            type_, subtype_ = kinds
            # allowing some user flexibility
            if subtype_ in ["cal"]:
                subtype_ = "calibrated"
            elif subtype_ in ["defcal"]:
                subtype_ = "defaultcalibrated"
            elif subtype_ in ["fluxcal"]:
                subtype_ = "invsensitivity"
                
            # parsing the input                
            if subtype_ !="*":
                regex  = PRODSTRUCT_RE[type_][subtype_]
            else:
                regex  = "|".join([subsreg
                    for subsreg in PRODSTRUCT_RE[type_].values()])
                
        else:
            raise TypeError("Enable to parse the given kind: %s"%kind)

    path = get_datapath(date)
    if target in ["*"]:
        target = None
    if extention in ["*",".*"]:
        extention = None

    # - Parsing the files
    return [path+f for f in os.listdir(get_datapath(date))
               if re.search(r'%s'%regex, f) and
                 (target is None or re.search(r'%s'%target, f)) and
                 (extention is None or f.endswith(extention))]

def get_cleaned_sedmcube(filename):
    """ get sky and flux calibrated cube """

    cube = get_sedmcube(filename)

    fluxcalfile = io.fetch_nearest_fluxcal(date, cube.filename)
    fluxcal = fluxcalibration.load_fluxcal_spectrum(fluxcalfile)

    cube.remove_sky()
    #cube.scale_by(fluxcal.data) # old version
    cube.scale_by( fluxcal.get_inversed_sensitivity(cube.header.get("AIRMASS", 1.1)), onraw=False )

    return cube
  
#########################
#                       #
#   Reading the DB      #
#                       #
#########################
def get_datapath(YYYYMMDD):
    """ Return the full path of the current date """
    return REDUXPATH+"/%s/"%YYYYMMDD

def fetch_header(date, target, kind="ccd.crr", getkey=None):
    """ Look for a night file (using get_night_files()) and returns its header 
    or a value from it stored with the given getkey."""
    datafile = get_night_files(date, kind, target=target)
    if len(datafile)==0:
        return None
    # Get the entire header
    if getkey is None:
        
        if len(datafile)==1:
            return getheader(datafile[0])
        return {d.split("/"):getheader(d) for d in datafile}
    # Or just a key value from it?
    else:
        from astropy.io.fits import getval
        if len(datafile)==1:
            return getval(datafile[0],getkey)

        return {d.split("/")[-1]:getval(d,getkey) for d in datafile}


def fetch_nearest_fluxcal(date=None, file=None, mjd=None, kind="spec.fluxcal"):
    """ Look for the fluxcal_*.fits file the closest (in time) from the given file and returns it. """
    if date is None:
        if mjd is not None:
            date = Time(mjd, format="mjd").datetime.isoformat().split("T")[0].replace("-","")
        elif file is not None:
            date = filename_to_date(file)
        else:
            raise ValueError("file is None, then date and/or mjd must be given. None here")
        
    filefluxcal = get_night_files(date, kind)
    
    if len(filefluxcal)==0:
        warnings.warn("No %s file for the night %s"%(kind, date))
        return None
    if len(filefluxcal)==1:
        warnings.warn("Only 1 file of kind %s for the night %s"%(kind, date))
        return filefluxcal[0]

    import numpy as np
    if mjd is not None:
        target_mjd_obs = mjd
    else:
        try:
            target_mjd_obs  = getval(file,"MJD_OBS")
        except KeyError:
            warnings.warn("No MJD_OBS keyword found, returning most recent file")
            return filefluxcal[-1]
        
    fluxcal_mjd_obs = [getval(f,"MJD_OBS") for f in filefluxcal]

    return filefluxcal[ np.argmin( np.abs( target_mjd_obs - np.asarray(fluxcal_mjd_obs) ) ) ]

def filename_to_id(filename):
    """ """
    return filename.split("/")[-1].split( header_to_date( getheader(filename) ))[-1][1:9]

def header_to_date( header, sep=""):
    """ returns the datetiume YYYYMMDD associated with the 'JD' from the header """
    datetime = Time(header["JD"], format="jd").datetime

    return sep.join(["%4s"%datetime.year, "%02d"%datetime.month, "%02d"%datetime.day])

def filename_to_time(filename):
    """ """
    date, hour, minut, sec = filename.split("_ifu")[-1].split("_")[:4]
    return Time("-".join([date[i:j] for i,j in [[0,4],[4,6],[6,8]]]) +" "+ ":".join([hour, minut, sec]))

def filename_to_date(filename, iso=False):
    """ """
    if iso:
        return filename_to_time(filename).datetime.isoformat().split("T")[0]
    return filename.split("_ifu")[1].split("_")[0]

def fetch_guider(date, filename, astrom=True, extinction=".fits"):
    """ fetch the guider data for the given filename. """
    print("DEPRECATED fetch_guider(date, filename) -> filename_to_guider(filename)")
    return filename_to_guider(filename, astrom=astrom, extinction=extinction)

def filename_to_guider(filename, astrom=True, extinction=".fits", nomd5=True):
    """ """
    fileinfo = parse_filename(filename)
    dirname = os.path.dirname(filename)
    key = "astrom" if astrom else "guider"
    return [os.path.join(dirname,l) for l in os.listdir( get_datapath(fileinfo["date"]))
                if fileinfo["sedmid"] in l and key in l and (l.endswith(('.fits', '.gz')) if nomd5 else l) ] 

    
def parse_filename(filename):
    """ """
    filename = filename.split(".")[0]
    if filename.startswith("crr"):
        crr, b, ifudate, hh, mm, ss, *targetname  = filename.split("_")
    else:
        _, crr, b, ifudate, hh, mm, ss, *targetname = filename.split("_")

    if len(targetname)==0:
        targetname = None
    else:
        targetname = ("-".join(targetname).replace(" ","")).split(".")[0]
        
    date = ifudate.replace("ifu","")
    mjd = Time(f"{date[:4]}-{date[4:6]}-{date[6:]}"+" "+f"{hh}:{mm}:{ss}", format="iso").mjd
    return {"date":date,
            "sedmid":f"{ifudate}_{hh}_{mm}_{ss}",
           "mjd":mjd, 
           "name":targetname}

def filename_to_background_name(filename):
    """ predefined structure for background naming """
    last = filename.split("/")[-1]
    return "".join([filename.split(last)[0],"bkgd_"+last])

def get_night_schedule(YYYYMMDD):
    """ Return the list of observations (the what.list) """
    schedule_file = glob(get_datapath(YYYYMMDD)+"what*")
    if len(schedule_file)==0:
        warnings.warn("No 'what list' for the given night ")
        return None
    return open(schedule_file[0]).read().splitlines()
        
def is_file_stdstars(filename):
    """ Tests if the 'OBJECT' entry of the file header is associated with a Standard star exposure. (True / False)
    None is returned if the header do not contain an 'OBJECT' entry 
    (see `is_stdstars`)
    Returns
    -------
    bool or None
    """
    from astropy.io.fits import getheader
    return is_stdstars( getheader(filename) )

def is_stdstars(header):
    """ Tests if the 'OBJECT' of the given header is associated with a Standard star exposure. (True / False)
    None is returned if the header do not contain an 'OBJECT' entry 

    Returns
    -------
    bool or None
    """
    obj = header.get("OBJECT",None)
    if obj is None:
        return None

    stdnames = ["STD","Feige", "Hitlner", "LTT"]
    return any([s_ in obj for s_ in stdnames])


#########################
#                       #
#   NIGHT SOLUTION      #
#                       #
#########################
# - Mapper
def load_nightly_mapper(YYYYMMDD, within_ccd_contours=True):
    """ High level object to do i,j<->x,y,lbda """
    from .mapping import Mapper
    
    tracematch = load_nightly_tracematch(YYYYMMDD)
    wsol       = load_nightly_wavesolution(YYYYMMDD)
    hgrid      = load_nightly_hexagonalgrid(YYYYMMDD)
    if within_ccd_contours:
        from .sedm import INDEX_CCD_CONTOURS
        indexes = tracematch.get_traces_within_polygon(INDEX_CCD_CONTOURS)
    else:
        indexes = list(wsol.wavesolutions.keys())
    
    mapper = Mapper(tracematch= tracematch, wavesolution = wsol, hexagrid=hgrid)
    mapper.derive_spaxel_mapping( indexes )
    return mapper

# - TraceMatch
def load_nightly_tracematch(YYYYMMDD, withmask=False):
    """ Load the spectral matcher.
    This object must have been created. 
    """
    from .spectralmatching import load_tracematcher
    if not withmask:
        return load_tracematcher(get_datapath(YYYYMMDD)+"%s_TraceMatch.pkl"%(YYYYMMDD))
    else:
        try:
            return load_tracematcher(get_datapath(YYYYMMDD)+"%s_TraceMatch_WithMasks.pkl"%(YYYYMMDD))
        except:
            warnings.warn("No TraceMatch_WithMasks found. returns the usual TraceMatch")
            return load_nightly_tracematch(YYYYMMDD, withmask=False)

# - HexaGrid
def load_nightly_hexagonalgrid(YYYYMMDD, download_it=True,
                                   nprocess_dl=1, **kwargs):
    """ Load the Grid id <-> QR<->XY position
    This object must have been created. 

    nprocess_dl: [int]
        Number of // download. 1 means no // processing


    """
    from .utils.hexagrid import load_hexprojection
    hexagrid_path = os.path.join(get_datapath(YYYYMMDD),"%s_HexaGrid.pkl"%(YYYYMMDD) )
    if os.path.isfile(hexagrid_path):
        return load_hexprojection( hexagrid_path )

    if not download_it:
        raise IOError(f"Cannot find an hexagrid for date {YYYYMMDD}")

    warnings.warn("cannot find the hexagrid for date {YYYYMMDD}, using ztquery.sedm to download it.")
    from ztfquery import sedm
    squery = sedm.SEDMQuery()
    squery.download_night_calibrations(YYYYMMDD, which="HexaGrid.pkl",
                                           nprocess=nprocess_dl, **kwargs)
    # has it been downloaded ?
    if os.path.isfile(hexagrid_path):
        return load_hexprojection( hexagrid_path )
    
    raise IOError(f"Cannot find an hexagrid for date {YYYYMMDD}, even after calling squery.download_night_calibrations()")
        
# - WaveSolution
def load_nightly_wavesolution(YYYYMMDD, subprocesses=False):
    """ Load the spectral matcher.
    This object must have been created. 
    """
    from .wavesolution import load_wavesolution
    if not subprocesses:
        return load_wavesolution(get_datapath(YYYYMMDD)+"%s_WaveSolution.pkl"%(YYYYMMDD))
    return [load_wavesolution(subwave) for subwave in glob(get_datapath(YYYYMMDD)+"%s_WaveSolution_range*.pkl"%(YYYYMMDD))]

# - 3D Flat
def load_nightly_flat(YYYYMMDD):
    """ Load the spectral matcher.
    This object must have been created. 
    """
    from pyifu.spectroscopy import load_slice
    return load_slice(get_datapath(YYYYMMDD)+"%s_Flat.fits"%(YYYYMMDD))

#########################
#                       #
#   PSF Product         #
#                       #
#########################
def get_psf_parameters(date, target=None, filepath=False):
    """ """
    path = get_datapath(date)
    if target == "*":
        target = None
    json_files =  [path+f for f in os.listdir(get_datapath(date))
                       if re.search(r'%s'%PRODSTRUCT_RE["psf"]["param"], f)
                       and (target is None or re.search(r'%s'%target, f))
                       and ("defcal") not in f]
    if filepath:
        return json_files
    import json
    return {f.split("/")[-1]:json.load( open(f) ) for f in json_files}
                    
def load_psf_param(date, keys_to_add=None):
    """ """
    from pyifu.adr import ten
    psfdata = get_psf_parameters(date)
    for s, d in psfdata.items():
        d["adr"]["delta_parangle"] = (d["adr"]["parangle"] - d["adr"]["parangle_ref"])%360
        d["adr"]["delta_parangle.err"] = d["adr"]["parangle.err"]
        sour =  s.split(date)[-1][:9]
        if keys_to_add is not None:
            d["header"] = {}
            for k in keys_to_add:
                if "DEC" in k or "RA" in k:
                    val = ten(fetch_header(date, sour, getkey=k))
                else:
                    val = fetch_header(date, sour, getkey=k)
                d["header"][k.lower()] = val
                
    return psfdata

#########################
#                       #
#   References          #
#                       #
#########################
def load_telluric_line(filter=None):
    """ return a TelluricSpectrum (child of pyifu Spectrum) containing the telluric emission line
    from Kitt Peak National Observatory.
    
    Source:
       Buton et al. 2013 (SNIFS, Buton, C., Copin, Y., Aldering, G., et al. 2013, A&A, 549, A8) 
       using data from
       Hinkle, K. H., Wallace, L., & Livingston, W. 2003, BAAS, 35, 1260.
       _Please cite both._
       
    Returns
    -------
    TelluricSpectrum (child of pyifu Spectrum)
    """
    from .utils.atmosphere import load_telluric_spectrum
    return load_telluric_spectrum(_PACKAGE_ROOT+"data/KPNO_lines.fits", filter=filter)

def get_bad_standard_exposures():
    """ Get the list of recorded standard star exposures that cannot be used for calibration. """
    return open(os.path.join(_PACKAGE_ROOT,"data/bad_standard.txt")).read().splitlines()

def get_noncalspec_standards():
    """ Get the list of STD observed by SEDm that are not calspec standards. """
    return open(os.path.join(_PACKAGE_ROOT,"data/noncalspec_standard.txt")).read().splitlines()

#########################3
#
#  OUTPUT PROD          #
#
#########################

def _saveout_forcepsf_(filecube, cube, cuberes=None, cubemodel=None,
                      cubefitted=None, spec=None, bkgd=None, extraction_type="Force 3DPSF extraction: Spectral Model",
                    mode="auto", spec_info="", fluxcal=True, nofig=False):
    # Cube Model
    if cubemodel is not None:
        cubemodel.set_header(cube.header)
        cubemodel.header["SOURCE"]   = (filecube.split("/")[-1], "This object has been derived from this file")
        cubemodel.header["PYSEDMT"]  = ("Force 3DPSF extraction: Model Cube", "This is the model cube of the PSF extract")
        cubemodel.header["EXTRTYPE"]  = (mode, "Kind of PSF extraction")
        cubemodel.writeto(filecube.replace(PROD_CUBEROOT,"forcepsfmodel_%s_"%mode+PROD_CUBEROOT))

    if cubefitted is not None:
        cubefitted.set_header(cube.header)
        cubefitted.header["SOURCE"]   = (filecube.split("/")[-1], "This object has been derived from this file")
        cubefitted.header["PYSEDMT"]  = ("Force 3DPSF extraction: Fitted Cube", "This is the model cube of the PSF extract")
        cubefitted.header["EXTRTYPE"]  = (mode, "Kind of PSF extraction")
        cubefitted.writeto(filecube.replace(PROD_CUBEROOT,"forcepsf_fitted_%s_"%mode+PROD_CUBEROOT))
        
    if cuberes is not None:
        # Cube Residual                
        cuberes.set_header(cube.header)
        cuberes.header["SOURCE"]   = (filecube.split("/")[-1], "This object has been derived from this file")
        cuberes.header["PYSEDMT"]  = ("Force 3DPSF extraction: Residual Cube", "This is the residual cube of the PSF extract")
        cuberes.header["EXTRTYPE"]  = (mode, "Kind of PSF extraction")
        cuberes.writeto(filecube.replace(PROD_CUBEROOT,"psfres_%s_"%mode+PROD_CUBEROOT))
    
    # ----------------- #
    # Save the Spectrum #
    # ----------------- #
    kind_key =""
    # - build the spectrum
    if spec is not None:
        for k,v in cube.header.items():
            if k not in spec.header:
                spec.header.set(k,v)

        spec.header["SOURCE"]   = (filecube.split("/")[-1], "This object has been derived from this file")
        spec.header["PYSEDMT"]  = (extraction_type, "This is the fitted flux spectrum")
        spec.header["EXTRTYPE"]  = (mode, "Kind of extraction")

        fileout = filecube.replace(PROD_CUBEROOT,PROD_SPECROOT+"%s_%s_"%(kind_key, mode+spec_info))
        spec.writeto(fileout)
        spec.writeto(fileout.replace(".fits",".txt"), ascii=True)
    
        spec._side_properties["filename"] = fileout

        if not nofig:
            from pyifu import get_spectrum
            spec_to_plot = get_spectrum(spec.lbda, spec.data,
                                        variance=spec.variance if spec.has_variance() else None,
                                        header=spec.header)
            spec_to_plot.show(savefile=spec.filename.replace(".fits", ".pdf"),
                              show_zero=fluxcal, show=False)
            spec_to_plot.show(savefile=spec.filename.replace(".fits", ".png"),
                              show_zero=fluxcal, show=False)
        
    # - background
    if bkgd is not None:
        bkgd.set_header(cube.header)
        bkgd.header["SOURCE"]   = (filecube.split("/")[-1], "This object has been derived from this file")
        bkgd.header["PYSEDMT"]  = (extraction_type, "This is the fitted flux spectrum")
        bkgd.header["EXTRTYPE"]  = (mode, "Kind of extraction")
    
        fileout = filecube.replace(PROD_CUBEROOT,PROD_SPECROOT+"%s_%s_bkgd"%(kind_key,mode+spec_info))
        bkgd.writeto(fileout)
        bkgd.writeto(fileout.replace(".fits",".txt"), ascii=True)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import division

from PIL import Image
import os
import io
import numpy as np
import matplotlib.pyplot as mpl




def get_image_row(images, interpolation=3, height="min"):
    """ combine PIL images in row from left to right  """
    
    if type(height) is str:
        height   = getattr(np, height)([img_.size[1] for img_ in images])
    elif type(height) in [int, float]:
        height = height
    else:
        raise TypeError("height should be a int or a string (numpy attribute like min/max/median)")

    imgs_comn = np.hstack([img_.resize( np.round(np.asarray(img_.size)*( height / img_.size[1] ) ).astype("int"), interpolation)
                           for img_ in images])
    return Image.fromarray(imgs_comn)

def get_image_column(images, interpolation=3, width="min"):
    """ combine PIL images in column from top to bottom  """
    if type(width) is str:
        width   = getattr(np, width)([img_.size[0] for img_ in images])
    elif type(width) in [int, float]:
        width = width
    else:
        raise TypeError("width should be a int or a string (numpy attribute like min/max/median)")

    imgs_comn = np.vstack([img_.resize( np.round((np.asarray(img_.size)*( width / img_.size[0] ) )).astype("int"), interpolation)
                           for img_ in images])
    return Image.fromarray(imgs_comn)

def get_buffer(size, text="", hline=None, vline=None, fontsize='medium',
               xy=[0.5,0.5], va="center",ha="center", barcolor="k",
               textprop={}, barprop={}, get_figure=False):
    """ Uses matplotlib to create a simple full size axis in which you can add text and horizonthal lines """
    import os
    import matplotlib.pyplot as mpl
    fig = mpl.figure(figsize=size)
    ax  = fig.add_axes([0,0,1,1])
    ax.axis('off')
    ax.text(xy[0],xy[1], text, va=va,ha=ha,fontsize=fontsize, **textprop)
    if hline is not None:
        for hline_ in np.atleast_1d(hline):
            ax.axhline(hline_, color=barcolor, **barprop )
            
    if vline is not None:
        for vline_ in np.atleast_1d(vline):
            ax.axvline(vline_, color=barcolor, **barprop )
    
    if get_figure:
        return fig
    buf = io.BytesIO()
    fig.savefig(buf, format="png")
    return Image.open(buf)



#! /usr/bin/env python
# -*- coding: utf-8 -*-


""" This library contains the high level functionnality to go from x,y,lbda (cube coords) <-> i,j (ccd-coords) """

import numpy as np
from propobject import BaseObject

from shapely.geometry import Point, LineString

from .utils.tools import is_arraylike


# = Pysedm Information = #
from .sedm import SEDMSPAXELS   as SPAXEL_SHAPE
from .sedm import SEDM_CCD_SIZE as CCD_SHAPE
INVERTED_LBDA_X = True


##########################
#                        #
#   The Mapping Class    #
#                        #
##########################
class Mapper( BaseObject ):
    """ """
    PROPERTIES = ["tracematch","hexagrid", "wavesolution"]
    DERIVED_PROPERTIES = ["spaxel_mapping","spaxelslice"]
    # ================= #
    #  Initialization   #
    # ================= #

    def __init__(self, tracematch=None, hexagrid=None, wavesolution=None):
        """ """
        if tracematch is not None:
            self.set_tracematch(tracematch)
        if hexagrid is not None:
            self.set_hexagrid(hexagrid)
        if wavesolution is not None:
            self.set_wavesolution(wavesolution)

    # - derived
    def derive_spaxel_mapping(self, traceindexes):
        """ """
        import pyifu
        xy = np.asarray(self.hexagrid.index_to_xy(self.hexagrid.ids_to_index(traceindexes))).T
        self._derived_properties["spaxel_mapping"] = {i:xy_ for i,xy_ in zip(traceindexes,xy)}
        
        self._derived_properties["spaxel_slice"] = \
          pyifu.get_slice( self.traceindexes, xy, spaxel_vertices= np.dot( self.hexagrid.grid_rotmatrix, SPAXEL_SHAPE.T ).T )
          
        self._derived_properties["spaxel_polygon"] = {i:p_ for i,p_ in zip(traceindexes,self.spaxel_slice.get_spaxel_polygon(remove_nan=True)) }
    # ================= #
    #  Methods          #
    # ================= #

    # ------------- #
    #   GETTER      #
    # ------------- #
    def get_lbda_ij(self, lbda, traceindexes=None):
        """ returns all i,j coordinates for the given wavelength

        Parameters
        ----------
        lbda: [float] 
            wavelength in Angstrom

        traceindexes: [list/array] -optional-
            list of traces you want to use. 
            If None all the traces in self.traceindexes will be used.

        Returns
        -------
        dictionary 
        [{traceindex:[i,j] ... }
        """
        if traceindexes is None:
            traceindexes = self.traceindexes

        return {traceindex:self.lbda_to_ij(lbda, traceindex) for traceindex in self.traceindexes}
    
    def get_ij(self, x, y, lbda):
        """ 
        Reminder: i,j are the CCD coordinates
                  q,r are the MLA hexagonal coordiates
                  x,y are the MLA coordinate (in spaxels)
                  traceindex is the unique ID of a spaxel on the ccd
        """
        traceindex = self.xy_to_traceindex(x,y)
        return self.lbda_to_ij(lbda, traceindex)
        
    def get_xylbda(self, i, j):
        """ 
        
        Reminder: i,j are the CCD coordinates
                  q,r are the MLA hexagonal coordiates
                  x,y are the MLA coordinate (in spaxels)
                  traceindex is the unique ID of a spaxel on the ccd
        """
        traceindex = self.ij_to_traceindex(i,j)
        lbda       = self.ij_to_lbda(i, traceindex)
        x,y        = self.traceindex_to_xy(traceindex)
        return np.asarray([x,y,lbda])

    def get_expected_j(self, i,j):
        """ fetch the traceindex associated to the i,j coordinates
        and then returns where the j is supposed to be if the trace where perfectly 
        aligned. 
        
        Reminder: i,j are the CCD coordinates
                  q,r are the MLA hexagonal coordiates
                  x,y are the MLA coordinate (in spaxels)
                  traceindex is the unique ID of a spaxel on the ccd
            
        """
        traceindex = self.ij_to_traceindex(i,j)
        if is_arraylike(traceindex):
            return [self.traceindexi_to_j(t_,i) for t_ in traceindex]
        
        return self.traceindexi_to_j(traceindex,i)[1]
            
    # ------------- #
    #  CONVERSION   #
    # ------------- #
    # .................... #
    #  lbda <-> i,j        #
    # .................... #
    def ij_to_lbda(self, i, traceindex):
        """ converts the i-th ccd pixel to a wavelength for a given traceindex """
        i_eff = (CCD_SHAPE[1]-1)-i if INVERTED_LBDA_X else i # -1 because starts at 0
        return self.wavesolution.pixels_to_lbda(i_eff, traceindex)
    
    def lbda_to_ij(self, lbda, traceindex):
        """ provide the central i,j coordinate for given traceindex at the given wavelength.
        Remark: j is the center of the trace for the column corresponding to the wavelength 
        """
        i =  self.wavesolution.lbda_to_pixels(lbda, traceindex)
        return self.traceindexi_to_j(traceindex, i, inverted=INVERTED_LBDA_X)
    
    def traceindexi_to_j(self, traceindex, i, maxlines=1e4, inverted=False):
        """ get thethe center of the trace `traceindex` for the column `i`"""
        if is_arraylike(i):
            return np.asarray([self.traceindexi_to_j(traceindex, i_) for _ in i])
        
        if i is None: return np.asarray([None,None])

        i_eff = (CCD_SHAPE[1]-1)-i if inverted else i # -1 because starts at 0
        try:
            return i_eff, np.mean(self.tracematch.trace_polygons[traceindex].intersection(LineString([[i_eff,0],[i_eff, maxlines]])), axis=0)[1]
        except:
            return i_eff, np.NaN

    # .................... #
    #  i,j <-> traceindex  #
    # .................... #
    def ij_to_traceindex(self, i, j):
        """ provide the traceindex associated to the pixel i, j.
        If the pixel is not in any spaxel, a None is set

        Reminder: i,j are the CCD coordinates
                  q,r are the MLA hexagonal coordiates
                  x,y are the MLA coordinate (in spaxels)
                  traceindex is the unique ID of a spaxel on the ccd

        Returns
        -------
        list
        """
        if is_arraylike(i):
            return [self.ij_to_traceindex(i_,j_) for i_,j_ in zip(i,j)]
        
        pij = Point(i,j)
        traceindex = [i_ for i_ in self.traceindexes if self.tracematch.trace_polygons[i_].contains(pij)]
        if len(traceindex) > 1:
            raise ValueError("the ccd-pixel coordinates %d,%d belongs to more than 1 spaxel-trace"%(i,j))

        return traceindex[0] if len(traceindex)==1 else None

    def traceindex_to_ij(self, traceindex):
        """ """
        return self.tracematch.get_trace_vertices(traceindex)
    
    # .................... #
    #  x,y <-> traceindex  #
    # .................... #
    def xy_to_traceindex(self, x, y):
        """ each spaxel has a unique x,y, location that 
        can be converted into a unique index (1d-array) entry.
        This tools enables to know what is the index corresponding to the given
        2D (x,y) position.
        

        Reminder: i,j are the CCD coordinates
                  q,r are the MLA hexagonal coordiates
                  x,y are the MLA coordinate (in spaxels)
                  traceindex is the unique ID of a spaxel on the ccd
                  
        Parameters
        ----------
        xy: [2d array]
            x and y position(s) in the following structure:
            [x,y] or [[x0,y0],[x1,y1]]

        Returns
        -------
        list of indexes 
        """
        # - Following set_spaxels_mapping, v are list
        if is_arraylike(x):
            return [self.xy_to_traceindex(x_,y_) for x_,y_ in zip(x,y)]
        
        pxy = Point(x,y)
        traceindex = [i_ for i_ in self.traceindexes if self._spaxel_polygon[i_].contains(pxy)]
        if len(traceindex) > 1:
            raise ValueError("the spaxel coordinates (x=%d,y=%d) belongs to more than 1 spaxel-trace"%(x,y))

        return traceindex[0] if len(traceindex)==1 else None
        
    def traceindex_to_xy(self, traceindex):
        """ Each spaxel has a unique index (1d-array) entry.
        This tools enables to know what is the 2D (x,y) position 
        of this index

        Reminder: i,j are the CCD coordinates
                  q,r are the MLA hexagonal coordiates
                  x,y are the MLA coordinate (in spaxels)
                  traceindex is the unique ID of a spaxel on the ccd

        Returns
        -------
        [int,int] (x,y)
        """
        if is_arraylike(traceindex):
            return np.asarray([self.traceindex_to_xy(index_) for index_ in traceindex])
        
        if traceindex in self.spaxel_mapping.keys():
            return self.spaxel_mapping[traceindex]
        return None, None
        
        
    # ------------- #
    #   SETTER      #
    # ------------- #
    # - set
    def set_hexagrid(self, hgrid):
        """ """
        self._properties["hexagrid"] = hgrid
        
    def set_tracematch(self, tmap):
        """ """
        self._properties["tracematch"] = tmap
        
    def set_wavesolution(self, wsol):
        """ """
        self._properties["wavesolution"] = wsol
        
    # ------------- #
    #   PLOTTER     #
    # ------------- #
    
    # ================= #
    #  Properties       #
    # ================= #
    # Sources
    @property
    def hexagrid(self):
        """ object containing the spaxel index <-> x,y coordinates conversion """
        return self._properties["hexagrid"]
    
    @property
    def tracematch(self):
        """ object containing the spaxel index <-> i,j coordinates conversion """
        return self._properties["tracematch"]
    

    @property
    def wavesolution(self):
        """ object containing the lbda<-> pixel coordinates """
        return self._properties["wavesolution"]

    # Derived
    @property
    def traceindexes(self):
        """ 'ID' (int) of the spaxels """
        return np.asarray(list(self.spaxel_mapping.keys()))

    @property
    def spaxel_slice(self):
        """ pyifu Slice using traceindex as data """
        if self._derived_properties["spaxel_slice"] is None:
            raise AttributeError("spaxel_slice not defined. See `derive_spaxel_mapping()` method")
        return self._derived_properties["spaxel_slice"]
    
    @property
    def spaxel_mapping(self):
        """ Dictionary containing the central coordinate of traceindex spaxels 
        {traceindex_i:[x_i, y_], ...} for the ith spaxel. """
        if self._derived_properties["spaxel_mapping"] is None:
            raise AttributeError("spaxel_mapping not defined. See `derive_spaxel_mapping()` method")
        return self._derived_properties["spaxel_mapping"]
    
    @property
    def _spaxel_polygon(self):
        """ list of Shapely Polygon derived by the `derive_spaxel_mapping()` method """
        if self._derived_properties["spaxel_polygon"] is None:
            raise AttributeError("spaxel_mapping not defined. See `derive_spaxel_mapping()` method")
        return self._derived_properties["spaxel_polygon"]




#! /usr/bin/env python
# -*- coding: utf-8 -*-


""" Scripts to build the calibrated cubes step by step """
from .. import io
from ..sedm import get_sedmcube

def flat_cubes(date, lbda_min=7000, lbda_max=9000, ref="dome"):
    """ """
    baseroot = io.CUBE_PROD_ROOTS["cube"]["root"]
    newroot  = io.CUBE_PROD_ROOTS["flat"]["root"]
    
    # -------------- #
    # The Reference  #
    # -------------- #
    reffile = io.get_night_cubes(date, kind="cube", target=ref)

    if len(reffile)==0:
        raise ValueError("No cube reference for target %s in night %s"%(ref,date))
    
    refcube  = get_sedmcube(reffile[0])
    flatfied = refcube.get_slice(lbda_min=lbda_min, lbda_max=lbda_max, usemean=True)
    print(flatfied.mean())
    # ----------------- #
    # Build flat cubes  #
    # ----------------- #
    def build_flat_cube(cubefile):
        cube_       = get_sedmcube(cubefile)
        print(cubefile)
        cube_.scale_by(flatfied)
        cube_.writeto(cube_.filename.replace(baseroot,newroot))
        
    cubefiles = io.get_night_cubes(date, kind="cube")
    print(cubefiles)
    return [build_flat_cube(cube_) for cube_ in cubefiles]
    
    
    
    



#! /usr/bin/env python
# -*- coding: utf-8 -*-

""" Access the rainbow camera images """
#
# This method is an a
#
import os
import numpy as np
from astropy import time
from astropy.io import fits
from . import io
RAINBOW_DATA_SOURCE = "/scr2/sedm/raw/"
SEDMPY_CODE_PATH = "/scr2/sedmdrp/sedmpy/"
# READOUT_NOISE       = 4


# ================== #
#  Main Function     #
# ================== #
def build_meta_ifu_guider(ifufile, outdir=None, solve_wcs=True, verbose=False):
    """ Higher level function. 
    It:
    1) fetches the guider images from the rainbow camera raw directory
    2) Merges them into one stacked_guider image
    3) Solves the astrometry on the stacked_guider image.

    Steps 1 and 2 are made using `build_stacked_guider()`
    Step 3 is made using `solve_astrometry` 

    Parameters
    ----------
    ifufile: [string]
        Path of a ifu .fits (or derived .fits as long as they contain
        the basic associated header keywords)
       
    outdir: [string] -optional-
        Where does the guider image should be saved. 
        If None, it will be in the same directory as the `infufile`
        
    solve_wcs: [bool] -optional-
        Shall "do_astro" (based on astrometry.net) be ran on the stacked image?

    verbose: [bool] -optional-
        Extra output
        
    Returns
    -------
    Void (creates a guider_`ifufile`)
    """
    savefile = build_stacked_guider(ifufile, outdir)
    if savefile:
        if verbose:
            print(" guider image built for %s" % ifufile)
            print(savefile)
        if solve_wcs:
            if verbose:
                print(" running astrometry on %s" % savefile)
            run_do_astrom(savefile)
            if not os.path.isfile(savefile.replace(".fits", "_astrom.fits")):
                print("do_astrom has failed.")
            else:
                print("do_astrom succeeded.")
    else:
        print("ERROR - unable to build guider image")


# ================== #
#   Function         #
# ================== #
def build_stacked_guider(ifufile, outdir=None, overwrite=True):
    """ 
    This function:
    1) fetches the guider images from the rainbow camera raw directory
       [using `get_ifu_guider_images`]
    2) Merges them into one stacked_guider image
       [using `get_ifu_guider_images`]


    Parameters
    ----------
    ifufile: [string]
        Path of a ifu .fits (or derived .fits as long as they contain
         the basic associated header keywords)
       
    outdir: [string] -optional-
        Where does the guider image should be saved. 
        If None, it will be in the same directory as the `infufile`

    overwrite: [bool] -optional-
        Set to overwrite existing file

    """
    guiders = get_ifu_guider_images(ifufile)
    stacked_image, nstack, avscl = stack_images(guiders)
    # did we get any?
    if nstack < 1:
        print("ERROR - no guider images found")
        return None
    # - building the .fits
    date = io.header_to_date(fits.getheader(ifufile))
    filein = ifufile.split("/")[-1]
    if outdir is None:
        outdir = io.get_datapath(date)
        
    savefile = outdir+"/guider_%s"%filein
    hdulist = fits.HDUList([fits.PrimaryHDU(stacked_image,
                                            fits.getheader(guiders[0]))])
    hdulist[0].header['NSTACK'] = nstack
    hdulist[0].header['STACKMTH'] = 'median'
    hdulist[0].header['SCALEMTH'] = 'median'
    hdulist[0].header['STACKSCL'] = avscl
    hdulist.writeto(savefile, overwrite=overwrite)
    return savefile

    
def run_do_astrom(guider_filename_fullpath):
    """ Run the do_astrom script in /scr2/sedmdrp/bin
    """
    try:
        do_astrom = os.path.join(os.environ['SEDMPY'], 'bin/do_astrom')
    except KeyError:
        do_astrom = SEDMPY_CODE_PATH + 'bin/do_astrom'
    import subprocess
    cmd = [do_astrom, guider_filename_fullpath]
    print(" ".join(cmd))
    subprocess.call(cmd)
    # Test results
    astrom_output = guider_filename_fullpath.replace(
        ".fits", "_astrom.fits").replace(".gz", "")
    if not os.path.exists(astrom_output):
        print("ERROR - astrometry failed, trying a median subtraction")
        from scipy import ndimage
        ff = fits.open(guider_filename_fullpath, mode='update')
        image = ff[0].data * 1.0    # Ensure we are float
        fsize = 15
        print("making median filter...")
        medfilt = ndimage.median_filter(image, fsize, mode='constant')
        ff[0].data = image - medfilt
        ff[0].header['MEDSUB'] = True, ' Median subtracted: %d px' % fsize
        ff.close()
        print("Done.  Re-doing do_astrom")
        subprocess.call(cmd)
    
    
# ================== #
#   Tools            #
# ================== #

def get_rainbow_datapath(ut_date):
    """ returns the path of the rainbow camera data """
    try:
        source = os.environ['SEDMRAWPATH']
    except KeyError:
        source = RAINBOW_DATA_SOURCE
    return os.path.join(source, ut_date) + "/"


def get_ifu_guider_images(ifufile):
    """ """    
    ifu_header = fits.getheader(ifufile)
    
    fileid = io.filename_to_id(ifufile)
    # - starting
    jd_ini = time.Time("%s %s" % (io.header_to_date(ifu_header, sep="-"),
                                  fileid.replace("_", ":"))).jd
    # - end
    jd_end = jd_ini + ifu_header['EXPTIME'] / (24.*3600)
    # - Where are the guider data  ?
    rb_dir = get_rainbow_datapath(io.header_to_date(ifu_header))
    # - Return them
    flist = os.listdir(rb_dir)
    rb_list = []
    for f in flist:
        # Use only *.fit* files
        if 'fit' not in f.split('/')[-1]:
            continue
        try:
            ff = fits.open(rb_dir+f)
        except OSError:
            print("WARNING - corrupt fits file: %s" % f)
            continue
        if "CAM_NAME" in ff[0].header:
            # Skip IFU images
            if "IFU" in ff[0].header['CAM_NAME']:
                continue
        else:
            if 'ifu' in f.split('/')[-1]:
                continue
        if "JD" in ff[0].header:
            # Images used to guide ifufile
            if jd_ini <= ff[0].header["JD"] <= jd_end:
                rb_list.append(rb_dir+f)
        else:
            print("WARNING - no JD keyword in %s" % f)
    return rb_list


def stack_images(rainbow_files, method="median", scale="median"):
    """ return a 2D image corresponding of the stack of the given data """
    # - 
    if scale not in ['median']:
        raise NotImplementedError("only median scaling implemented (not %s)"
                                  % scale)
    
    if method not in ['median']:
        raise NotImplementedError(
            "only median stacking method implemented (not %s)" % method)
    # Load the normalized data

    datas = []
    scales = []
    scaling = 1.
    avscale = 1.
    for f_ in rainbow_files:
        data_ = fits.getdata(f_)
        # header_= fits.getheader(f_) all have the same gain and readout noise
        if scale in ["median"]: 
            scaling = np.median(data_)
            scales.append(scaling)
        datas.append(data_/scaling)
        avscale = np.mean(scales)

    return np.median(datas, axis=0)*avscale, len(datas), avscale



#! /usr/bin/env python
# -*- coding: utf-8 -*-



#################################
#
#   MAIN
#
#################################
if  __name__ == "__main__":

    import argparse
    import numpy as np

    from pysedm import get_sedmcube, io



    # ================= #
    #   Options         #
    # ================= #
    parser = argparse.ArgumentParser(
        description=""" run the interactive plotting of a given cube""",
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('infile', type=str, default="None",
                        help='cube filepath')

    # // AUTOMATIC EXTRACTION
    parser.add_argument('--inputcube',   action="store_true", default=False,
                        help='Is the input a cube ?')

    parser.add_argument('--reducer',  type=str, default=io.SEDM_REDUCER,
                        help='What is your name? This will be added in the header [default: $SEDM_REDUCER]')

    #  which extraction
    parser.add_argument('--aperture',  type=str, default=None,
                        help='Do you want a simple aperture extraction ?')

    parser.add_argument('--ap_bkgdscale',  type=float, default=1.4,
                        help='Default width scale of the outter background ring [default 1.4]')

    # Auto PSF
    parser.add_argument('--auto',  type=str, default=None,
                        help='Shall this run an automatic PSF extraction')

    parser.add_argument('--autorange',  type=str, default="4000,8000",
                        help='Wavelength range [in Angstrom] for measuring the metaslice PSF')

    parser.add_argument('--autobins',  type=int, default=7,
                        help='Number of bins within the wavelength range (see --autorange)')

    parser.add_argument('--buffer',  type=float, default=10,
                        help='Radius [in spaxels] of the aperture used for the PSF fit. (see --centroid for aperture center)')



    parser.add_argument('--psfmodel',  type=str, default="NormalMoffatTilted",
                        help='PSF model used for the PSF fit: NormalMoffat{Flat/Tilted/Curved}')

    parser.add_argument('--seeing',  type=float, default=2,
                        help='NOT READY YET Expected seeing (not quite in arcsec). This helps the PSF model to converge.')

    # Contour separation with contsep.py
    parser.add_argument('--contsep',  action="store_true", default=False,
                        help='Shall this run contsep to avoid non-target spaxels within the fit.')

    parser.add_argument('--contsep_offset', nargs=2, type=float, default=[0, 0],
                        help='Offset between reference image and cube image for the contsep module.')

    parser.add_argument('--contsep_fakemag', type=float, default=16.0,
                        help="Fake magnitude of the target for a reference image.")

    parser.add_argument('--contsep_forcedmag', type=float, default=0.0,
                        help="Addition to contsep_mag to select more host spaxels.")

    # Cosmic ray removal with byecr.py
    parser.add_argument('--byecr', action="store_true", default=False,
                         help="Shall this run byecr to remove cosmic rays at a specific spaxel at a specific wavelengh.")

    parser.add_argument('--byecr_lbda', type=float, default=None,
                         help="To check cosmic rays at a given wavelength.")

    parser.add_argument('--byecr_cut', type=float, default=5.0,
                         help="Cut criteria for byecr module.")

    parser.add_argument('--byecr_wspectral', action="store_true", default=False,
                         help="Also use a spectral filtering for byecr.")

    parser.add_argument('--byecr_showcube', action="store_true", default=False,
                         help="Show a cube with detected cosmic rays.")


    # Centroid
    parser.add_argument('--centroid',  type=str, default="auto", nargs="+",
                        help='Where is the point source expected to be ?'+
                            "\nYou have 3 potential cases:"+
                            "\n   * --centroid auto: Use astrometry from guider if possible or falls back to the 'brightest' mode"+
                            "\n   * --centroid brightest: Use the position of the bightest spaxels "+
                            "\n   * --centroid x0,y0: Use the given x0 y0 position (separated by a coma)."
                            )

    parser.add_argument('--centroiderr',  type=str, default="None",nargs=2,
                        help='Where is the point source location guess boundaries ?'+
                        "\nYou have 2 potential cases:"+
                            "\n   * --centroiderr None: [3 3] if position from the guider, [5 5] if from the 'brightest' mode "+
                            "\n   * --centroiderr [X Y]: use the given values")

    # wavelength step
    parser.add_argument('--display',  action="store_true", default=False,
                        help='Select the area to fit using the display function.')

    parser.add_argument('--vmin',  type=str, default="2",
                        help='Data Percentage used for imshow "vmin" when using the --display mode')

    parser.add_argument('--vmax',  type=str, default="98",
                        help='Data Percentage used for imshow "vmax" when using the --display mode')

    # Flux calibration
    parser.add_argument('--nofluxcal',  action="store_true", default=False,
                        help='No Flux calibration')

    parser.add_argument('--fluxcalsource',  type=str, default="None",
                        help="Path to a 'fluxcal' .fits file. This file will be used for the flux calibration"+
                        "\nIf nothing given, the nearest (in time and within the night) fluxcal will be used.")

    # wavelength step
    parser.add_argument('--lstep',  type=int, default=1,
                        help='Slice width in lbda step: default is 1, use 2 for fainter source and maybe 3 for really faint target')

    # I/O options
    parser.add_argument('--tag',  type=str, default="None",
                        help='Add the tag on output filename.')

    parser.add_argument('--nofig',    action="store_true", default=False,
                        help='Do not create any plot')

    parser.add_argument('--notroncation',  action="store_true", default=False,
                        help='Do not apply the blue and red end edge troncation on flux calibrated spectra')

    # - Standard Star object
    parser.add_argument('--std',  action="store_true", default=False,
                        help='Set this to True to tell the program you what to build a calibration spectrum from this object')






    args = parser.parse_args()


    # ================= #
    #   The Scripts     #
    # ================= #

    # --------- #
    #  Date     #
    # --------- #
    if not args.inputcube:
        date = args.infile

    # --------- #
    #  Parsing  #
    # --------- #
    if len(args.centroid)==1:
        args.centroid = args.centroid[0]
    elif len(args.centroid)>2:
        raise ValueError("--centroid should be 'auto', 'brightness' or X Y")




    # ------------- #
    #               #
    # APERTURE      #
    #               #
    # ------------- #
    if args.aperture is not None and len(args.aperture) >0:
        raise NotImplementedError("--aperture option not implemented yet. Use --auto")
        if args.std:
            raise NotImplementedError("--std option cannot be used (yet) with --aperture. Use --auto")



    # ------------- #
    #               #
    # 3D PSF        #
    #               #
    # ------------- #
    if args.auto is not None and len(args.auto) >0:

        for target in args.auto.split(","):

            filecubes = io.get_night_files(date, "cube.*", target=target.replace(".fits","")) if not args.inputcube else [target]
            print("cube file from which the spectra will be extracted [auto]: "+ ", ".join(filecubes))

            # - loop over the file cube
            for filecube in filecubes:
                print("Automatic extraction of target %s, file: %s"%(target, filecube))
                # ===================== #
                # Running ExtractStar   #
                # ===================== #
                #
                # DATA IN
                #
                cube = get_sedmcube(filecube)
                cube.header['REDUCER'] = args.reducer
                #
                # OPTIONS
                #
                es_options = {
                    # Display mode ?
                    "display":args.display, "displayprop":{"vmin":args.vmin, "vmax":args.vmax},

                    # Metaslices
                    # -- How many slices
                    "step1range": np.asarray(args.autorange.split(","), dtype="float"),
                    "step1bins": args.autobins,
                    # -- Centroid guess
                    "centroid":args.centroid, "prop_position":{"centroiderr":args.centroid},
                    # Sub IFU
                    "spaxelbuffer": args.buffer,
                    # PSF Model
                    "psfmodel": args.psfmodel,
                    "spaxels_to_use":None, # not ready yet, provide here directly the spaxels to be used
                    "spaxels_to_avoid":None, # not ready yet, provide here directly the spaxels to be used
                    "fwhm_guess": args.seeing,
                    # Spectral options
                    "slice_width": int(args.lstep),
                    #
                    #
                    "verbose":True,
                    }

                # This means that you want to remove non-target spaxel contamination
                if args.contsep:
                    from pysedm import contsep
                    print(" Starting contsep non-target spaxel removal ".center(50, "-"))
                    targetid = io.filename_to_id(filecube)
                    cont = contsep.get_spaxels_from_constsep(date, targetid,
                                                             offset=args.contsep_offset,
                                                             fake_mag=args.contsep_fakemag,
                                                             forced_mag=args.contsep_forcedmag)
                    es_options["spaxels_to_avoid"] =  list( cont.get_others_spaxels(spaxels_id=False) )

                # Remove cosmic ray at a specific spaxel at a given wavelength.
                if args.byecr:
                    from pysedm import byecr
                    print(" Starting byecr cosmic ray removal ".center(50, "-"))
                    targetid = io.filename_to_id(filecube)
                    byecrclass = byecr.get_cr_spaxels_from_byecr(date, targetid)
                    cr_df = byecrclass.get_cr_spaxel_info(lbda_index=args.byecr_lbda,
                                                          cut_criteria=args.byecr_cut,
                                                          wspectral=args.byecr_wspectral)

                    cube.data[cr_df["cr_lbda_index"], cr_df["cr_spaxel_index"]] = np.nan

                    print("Number of detected cosmic rays = %i" %len(cr_df))
                    print("Number of cosmic ray affected spaxels = %i" %len(np.unique(cr_df["cr_spaxel_index"])))


                # ===================== #
                # SOURCE EXTRACTION     #
                # ===================== #
                print(" Starting extract_pointsource ".center(50, "-"))
                try:
                    es_out = cube.extract_pointsource(**es_options)
                except IOError:
                    # Astrometry was not solved
                    print("ERROR: Astrometry file not found, "
                          "using brightest spaxel instead")
                    es_options["centroid"] = "brightest"
                    es_out = cube.extract_pointsource(**es_options)
                except AttributeError:
                    # Astrometry solved, but target outside IFU
                    print("ERROR: Astrometry places target outside IFU, "
                          "using brightest spaxel instead")
                    es_options["centroid"] = "brightest"
                    es_out = cube.extract_pointsource(**es_options)
                    cube.header['QUALITY'] = 3
                # -> The output object
                es_object = cube.extractstar

                if args.contsep:
                    es_object.raw_spectrum.header.set("CONTMAG", cont.target_contsep_mag,
                                                          "contsep splitting magnitude")
                    es_object.raw_spectrum.header.set("NCONTSPX",
                                                len(es_options["spaxels_to_avoid"]),
                                                "total number of non-target spaxels from contsep")

                if args.byecr:
                    es_object.raw_spectrum.header.set("NCR", len(cr_df),
                                                "total number of detected cosmic-rays from byecr")
                    es_object.raw_spectrum.header.set("NCRSPX",
                                                len(np.unique(cr_df["cr_spaxel_index"])),
                                                "total number of cosmic-ray affected spaxels")

                # -
                # - SAVING
                # -
                add_tag = "_%s" % args.tag if args.tag is not None and args.tag not in ["None", ""] else ""
                add_info_spec = "_notfluxcal" if not es_object.is_spectrum_fluxcalibrated() else ""
                spec_info = "_lstep%s" % es_options["slice_width"] + add_info_spec

                print("Tag added", add_tag)
                es_object.writeto(basename=None, add_tag="auto" + add_tag, add_info=None)

                # -
                # - PLOTTING
                # -
                plot_tag = "_auto" + add_tag + spec_info + "_"
                if not args.nofig:
                    print(" Saving figures ".center(50, "-"))
                    es_object.show_adr(            savefile=es_object.basename.replace("{placeholder}","adr_fit" + plot_tag))
                    es_object.show_metaslices(     savefile=es_object.basename.replace("{placeholder}","metaslices" + plot_tag))
                    es_object.show_extracted_spec( savefile=es_object.basename.replace("{placeholder}","extracted_spec" + plot_tag))
                    es_object.show_mla(            savefile=es_object.basename.replace("{placeholder}","ifu_spaxels_source" + plot_tag))
                    es_object.show_psf(            savefile=es_object.basename.replace("{placeholder}","psfprofile" + plot_tag), sliceid=2)
                    es_object.spectrum.show(       savefile=es_object.basename.replace("{placeholder}","spec" + plot_tag))
                    if args.contsep:
                        cont.show_ifudata( wcontour=False, wtargetspaxel=True, wotherspaxel=True,
                        savefile=es_object.basename.replace("{placeholder}","contsep" + plot_tag) )
                    if args.byecr_showcube:
                        byecrclass.show_cr_spaxels( wcube=args.byecr_showcube, lbda_index=args.byecr_lbda, cut_criteria=args.byecr_cut, wspectral=args.byecr_wspectral,
                        savefile=es_object.basename.replace("{placeholder}","byecr" + plot_tag) )

                # -
                # - Standard Specific
                # -
                if args.std:
                    print("Standard Star object, starting estimation of flux calibration")
                    #
                    # Flux Calibrator
                    #
                    speccal, fl = es_object.get_fluxcalibrator(persecond=True, troncate_edges=False)
                    if speccal is None:
                        print("WARNING: Getting the Flux Calibrator Failed.")
                    else:
                        speccal.writeto(speccal.filename.replace("fluxcal", "fluxcal" + plot_tag))

                    #
                    # PLOTTING
                    #
                    if not args.nofig:
                        if cube.header['IMGTYPE'].lower() in ['standard'] and es_object.is_spectrum_fluxcalibrated():
                            from pysedm.fluxcalibration import show_fluxcalibrated_standard
                            show_fluxcalibrated_standard(es_object.spectrum, savefile=es_object.basename.replace("{placeholder}","calibcheck_spec" + plot_tag)+".pdf")
                            show_fluxcalibrated_standard(es_object.spectrum, savefile=es_object.basename.replace("{placeholder}","calibcheck_spec" + plot_tag)+".png")
                        if fl is not None:
                            fl.show(savefile=speccal.filename.replace(".fits",".pdf"),
                                        show=False)
                            fl.show(savefile=speccal.filename.replace(".fits",".png"),
                                        show=False)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

""" This modules contains the Wavelength solution tools. """

import warnings
import numpy as np
import matplotlib.pyplot as mpl
from scipy         import optimize
from astropy.utils.console import ProgressBar

try:
    from pynverse import inversefunc
    _HASPYNVERSE = True
except ImportError:
    warnings.warn("Cannont Import pynverse, limited wavelength solution tools | pip install pynverse")
    _HASPYNVERSE = False
# - External Modules
from propobject              import BaseObject
from pyifu.spectroscopy      import Spectrum
import pyifu
import modefit

# - Internal Modules
from .ccd import CCD

# Vacuum wavelength
# from KECK https://www2.keck.hawaii.edu/inst/lris/arc_calibrations.html
# Cadmium http://www.avantes.com/images/stories/applications/pagina_93-94_cadmium.jpg
# Xenon from SNIFS
#
#
# http://www.astrosurf.com/buil/us/spe2/hresol4.htm
#

# From Keck: Hg
HG_LINES = {"air":[3125.67 , 3131.70 , 3341.48 , 3650.153,
                  3654.84 , 3663.279, 4046.563, 4077.831,
                  4358.327, 4916.068, 5460.750, 5769.598,
                  5790.663],
            "vacuum":  [3126.580, 3132.463, 3342.445, 3651.198,
                        3655.883, 3664.327, 4047.708, 4078.988,  
                        4359.560, 4917.440, 5462.268, 5771.210,
                        5792.276],
             "vacuum_blended": [3129.44,3345.04, 3650.87, 3663.18] 
            }
    
CD_LINES = {"air" : [3080.822, 3082.593, 3133.167, 3252.524, 
                     3261.055, 3403.652, 3466.200, 3467.655, 
                     3499.952, 3610.508, 3612.873, 3614.453, 
                     3649.558, 3981.926, 4140.500, 4306.672, 
                     4412.989, 4662.352, 4678.149, 4799.912, 
                     5085.822, 5154.660, 6099.142, 6111.490, 
                     6325.166, 6330.013, 6438.470, 6778.116, 
                     7345.670],
                     
          "vacuum" : [3081.71730, 3083.48870, 3134.07460, 3253.46220, 
                      3261.99510, 3404.62880, 3467.19230, 3468.64780, 
                      3500.95320, 3611.53750, 3613.90330, 3615.48370, 
                      3650.59830, 3983.05200, 4141.46650, 4307.88320, 
                      4414.22880, 4663.65720, 4679.45870, 4801.25400, 
                      5087.23930, 5156.09640, 6100.83050, 6113.18700, 
                      6326.91510, 6331.76360, 6440.24900, 6779.98650, 
                      7347.69410],
           "vacuum_blended": [3466.85, 3611.01, 4679.325] 
                }


XE_LINES = {"air": [ 7119.614 ,  7642.0124,  8231.6305,  8231.6305,  8280.1113,
                     8280.1113,  8346.8112,  8347.2311,  8408.2036,  8409.1843,
                     8819.4029,  8952.2468,  9045.4415,  9162.6396,  9513.3743,
                     9513.3743,  9685.3175,  9718.1586,  9799.6964],
                     
            # vacuum DERIVED inversing var_to_air_sdss, see utils.tools.fetch_vac
            "vacuum": [7121.58, 7644.12, 8233.90, 8233.90,
                       8282.39, 8282.39, 8349.11, 8349.53,
                       8410.52, 8411.50, 8821.83, 8954.71,
                       9047.93, 9165.16, 9515.99, 9515.99,
                       9687.98, 9720.83, 9802.39],
            "vacuum_blended": [8411.0] 
                }


# -------------- #
#  SEDM          #
# -------------- #
REFWAVELENGTH = 7000

# For Information:
#    Sodium skyline fitted by SNIFS is at 5892.346 Angstrom
SODIUM_SKYLINE_LBDA = 5900#5896
TELLURIC_REF_LBDA   = 7630 #  7624.5

_REFORIGIN = 69

LINES= {"Hg": # IN VACUUM
            { 
               np.mean([5771.210, 5792.276])   : {"ampl":13. ,"mu":201-_REFORIGIN,
                           "doublet":False,
                        "info":"merge of 5771.210, 5792.276 blended"},
               5462.268 : {"ampl":62.,"mu":187-_REFORIGIN},
               4359.560   : {"ampl":50. ,"mu":127-_REFORIGIN},
               4047.708 : {"ampl":10. ,"mu":105-_REFORIGIN}, 
              
               },
        "Cd":  # IN VACUUM
              {4679.325 : {"ampl":14. ,"mu":146-_REFORIGIN}, 
               4801.254 : {"ampl":40. ,"mu":153-_REFORIGIN},
               5087.239 : {"ampl":60. ,"mu":169-_REFORIGIN},
               6440.249 : {"ampl":19. ,"mu":227-_REFORIGIN}, 
               # - Cd and Xe seems to have the same lines
               # Almost same wavelength but diffent enough to save the code
               8280.01  : {"ampl": 1. ,"mu":275-_REFORIGIN,
                            "backup":"Xe", # used only if Xe is not there
                            "doublet":True,"info":"merge of 8230, 8341"},
               #8818.51  : {"ampl": 0.8,"mu":295-_REFORIGIN},
               #9000.01  : {"ampl": 0.5,"mu":302-_REFORIGIN,
               #             "doublet":True,"info":"merge of 8945, 9050"},
                },
        # Important for Xe. Keep the Bright line, the First one. or change get_line_shift()
#        "Xe": {8280.    : {"ampl": 1. ,"mu":280-_REFORIGIN,
#                            "doublet":True,"info":"merge of 8230, 8341"},
#               8818.5   : {"ampl": 0.8,"mu":295-_REFORIGIN},
#               9000     : {"ampl": 0.5,"mu":302-_REFORIGIN,
#                            "doublet":True,"info":"merge of 8945, 9050"},
#               }

        "Xe": {np.average([8233.90, 8282.39], weights=[2,1])  : {"ampl": 16. ,"mu":280-_REFORIGIN,
                            "doublet":False},# yes but really close
                            
               np.average([8349.11, 8411.00], weights=[2,3])   : {"ampl": 3. ,"mu":283-_REFORIGIN,
                            "doublet":False}, # yes but really close
                            
                            
               8821.83    : {"ampl": 11.,"mu":294-_REFORIGIN},
                   
               np.average([8954.71,9047.93]) : {"ampl": 11.,"mu":298-_REFORIGIN,
                             "doublet":True , "info": "merge of lines 9854.71,9047.93"},
               9165.16    : {"ampl": 4.5,"mu":303-_REFORIGIN},
               
               # small lines but isolated
               7644.12    : {"ampl": 1.,"mu":264-_REFORIGIN},
               
               #9425.    : {"ampl": 2.,"mu":309-_REFORIGIN,
               #                "doublet":True , "info": "merge of lines 9400 and 9450"},
               
               #9802.8    : {"ampl": 0.4,"mu":317-_REFORIGIN},
               #9938.0    : {"ampl": 0.4,"mu":320-_REFORIGIN},
               
               }
        }

###########################
#                         #
#  Generators             #
#                         #
###########################
def load_wavesolution(filenames):
    """ Load the wavesolution from saved data """
    wsol = WaveSolution()
    wsol.load(filenames)
    return wsol
    
def get_wavesolution(*lamps):
    """ Loads the object that enables to build the WaveSolution. """
    wsol = WaveSolution()
    [wsol.add_lampccd(l) for l in lamps]
    return wsol

def get_arcspectrum(x, y, databound, dy=None, name=None):
    """ """
    spec = ArcSpectrum(None)
    spec.create(lbda=x, data=y, variance=dy**2 if dy is not None else None, logwave=False)
    spec.set_arcname(name)
    spec.set_databounds(*databound)
    return spec

def get_arccollection(traceindex, lamps):
    """ """
    sol_ = ArcSpectrumCollection()
    
    for lamp in lamps:
        spec_ = lamp.get_spectrum(traceindex, on="data")[::-1]
        lbda_ = np.arange(len(spec_))
        sol_.add_arcspectrum( get_arcspectrum(x=lbda_, y=spec_,
                                            databound= np.sort(len(spec_) - lamp.tracematch.get_trace_xbounds(traceindex)),
                                            name=lamp.objname))
    sol_.set_databounds(*np.sort(len(spec_) - lamp.tracematch.get_trace_xbounds(traceindex)))
    return sol_

# ==================== #
#   Subprocessing      #
# ==================== #
def merge_wavesolutions(wavesolutions):
    """ """
    wsol = WaveSolution()
    for w in wavesolutions:
        for i in w.traceindexes:
            wsol.add_trace_wavesolution(i,w.wavesolutions[i])
        
    return wsol

# -------------------- #
#   MultiProcessing    #
# -------------------- #
def fit_spaxel_wavelesolution(arccollection, sequential=True,
                              contdegree=3, wavedegree=5,
                              saveplot=None, show=False,
                        plotprop={}):
    """ Fit the wavelength solution of the given trace.
    
    This matching is made in two steps:
    1) fit the central positions of the loaded arcspectrum emission lines.
       This fit is based on linear combination of gaussian lines + polynomial continuum.
    2) Fit a polynomial degree to modelize the relation between the 
       fitted central positions (c.f. step 1) with the corresponding wavelengths
       of these lines.

    Parameters
    ----------
    
    sequential: [bool] -optional-
        How the gaussian line profiles (central positions) should be fitted:
        - False: All at once (as if it was a unique spectrum)
        - True:  Once at the time (each arcspectrum fitted independently)

    contdegree: [int] -optional-
        The degree of the polynom underlying the gaussian lines
            
    wavedegree: [int] -optional-
        The degree of the polynom modelizing the pixel<->wavelength relation
            
    saveplot: [None or string] -optional-
        Save the wavelength solution figure. If None, nothing will be saved.
            
    show: [bool] -optional-
        Should the wavelength solution figure be shown?
        
    Returns
    -------
    None
    """
    # 3s 
    arccollection.fit_lineposition(contdegree=contdegree, sequential=sequential)
    # 0.01s
    arccollection.fit_wavelengthsolution(wavedegree, legendre=False)
    return arccollection

def fit_wavesolution(lamps, indexes, multiprocess=True):
    """ """
    from .utils import tools
    notebook = tools.is_running_from_notebook()

    if notebook:
        print("RUNNING FROM NOTEBOOK")
        try:
            return _fit_background_notebook_(lamps, indexes, multiprocess=True)
        except:
            warnings.warn("FAILING fit_background for notebooks")
            
    # Running from ipython/python    
    arccollections = [get_arccollection(id_, lamps) for id_ in indexes]
    
    wsolutions = ProgressBar.map(fit_spaxel_wavelesolution,
                                arccollections,
                                multiprocess=multiprocess, step=2)
    #  these are the wavesolutions
    return {i_:wsol_.data for i_,wsol_ in zip(indexes, contval)}

def _fit_wavesolution_notebook_(lamps, indexes, multiprocess=True):
    """ """
    arccollections = [get_arccollection(id_, lamps) for id_ in indexes]
    
    # - Multiprocessing 
    if multiprocess:
        bar = ProgressBar( len(indexes), ipython_widget=True)
        import multiprocessing
        p = multiprocessing.Pool()
        res = {}
        for j, result in enumerate( p.imap(fit_spaxel_wavelesolution, arccollections)):
            res[indexes[j]] = result
            bar.update(j)
        bar.update(len(indexes))
        return res
    
    # - No multiprocessing 
    return {indexes[i_]: fit_spaxel_wavelesolution(arccollection)
                for i_,arccollection in enumerate(arccollections)}


###########################
#                         #
#  Flexure Correction     #
#                         #
###########################
class Flexure( BaseObject ):
    """ """
    PROPERTIES = ["cube", "mapper", "reference"]
    SIDE_PROPERTIES = ["spec_sodium", "spec_sodium_indexes",
                        "spec_telluric"]
    DERIVED_PROPERTIES = ["fit_telluric", "fit_sodium"]

    def __init__(self, cube, mapper=None):
        """ """
        self.set_cube(cube)
        self.set_mapper(mapper)
    # ================== #
    #   Method           #
    # ================== #
    # -------- #
    #  SHOW    #
    # -------- #
    def show(self, savefile=None, show=True):
        """ """
        
        fig = mpl.figure(figsize=[5,3])
        axna  = fig.add_axes([0.15,0.2,0.75,0.31])
        axte  = fig.add_axes([0.15,0.60,0.75,0.31])

        refcolor = "C1"
        # Telluric
        _ = [f_.show(color="C0", alpha=0.2, ecolor="C0", modelcolor="0.8",ax=axte, show=False) for f_ in self.fit_telluric]
        _ = [axte.axvline(f_.fitvalues["mu0"], color="C0", alpha=0.2, zorder=7) for f_ in self.fit_telluric]
        axte.axvline(TELLURIC_REF_LBDA, color=refcolor, zorder=8)

        _ = [f_.show(color="C0", alpha=0.2, ecolor="C0", modelcolor="0.8",ax=axna, show=False) for f_ in self.fit_sodium]
        _ = [axna.axvline(f_.fitvalues["mu0"], color="C0", alpha=0.2, zorder=7) for f_ in self.fit_sodium]
        axna.axvline(SODIUM_SKYLINE_LBDA, color=refcolor, zorder=8)

        for ax_ in [axte, axna]:
            ax_.spines['top'].set_visible(False)
            ax_.spines['right'].set_visible(False)
            ax_.spines['left'].set_visible(False)
            ax_.tick_params(labelsize="small")
            ax_.tick_params(axis="y",color="0.5", labelcolor="0.5", labelsize="x-small")

        axna.set_xlabel(r"Wavelength [$\AA$]", fontsize="medium")
        axte.set_ylabel(r"Flux [a.u]", fontsize="small", color="0.5")
        axna.set_ylabel(r"Flux [a.u]", fontsize="small", color="0.5")
        
        #
        # Info
        #
        prop_box = {'facecolor':'w', 'alpha':0.7, "edgecolor":"None"}
        for ax_,which_ in zip([axna,axte], ["sodium", "telluric"]):
            mu,muerr = self.get_line_position(which_)
            ax_.text(0,0.98, r"$\Delta \lambda=%.1f \pm %.1f\,\AA$"%(mu-self.reference[which_],muerr),
                 transform=ax_.transAxes, va="top", ha="left", fontsize="small", color="0.5",
                        bbox=prop_box)
            ax_.text(1,0, r"%s $%.1f\,\AA$"%(which_,self.reference[which_]),
                         transform=ax_.transAxes, fontsize="small", va="bottom", ha="right",
                         bbox=prop_box)
            
        # Generic information
        fig.text(0.5,0.98,"corresponds to a typical ccd 'i'-shift of %.2f pixels"%(self.get_i_flexure()),
                            va="top",ha="center", color="k",
                            bbox={'facecolor':'w', 'alpha':0.9,'edgecolor':'0.5', "boxstyle":"square"}, fontsize="small")
        if savefile is not None:
            fig.savefig(savefile)
        elif show:
            fig.show()
            
    # -------- #
    #  GETTER  #
    # -------- #
    def get_line_position(self, which="sodium", as_array=False, method="nanmean"):
        """ 
        The error is based on median statistics as long as there are more than 10 spectrum to measure the dispersion
        """
        from astropy.stats import mad_std
        mu, muerr = np.asarray([[t_.fitvalues["mu0"],t_.fitvalues["mu0.err"]] for t_ in eval("self.fit_%s"%which)]).T
        if as_array:
            return mu, muerr
        if method in ["wmean", "weightedmean"]:
            mean_ = np.average(mu, weights=1/muerr**2)
        elif method in ["mean", "nanmean", "median", "nanmedian"]:
            mean_ = eval("np.%s(mu)"%method)
        else:  
            raise ValueError("Unknown averaging method %"%method)
        
        err = mad_std(mu)/np.sqrt(len(mu)-1) if len(mu)>10 else np.std(mu)/np.sqrt(len(mu)-1)
        return mean_, err
        
    def get_i_flexure(self, use=["sodium", "telluric"], **kwargs):
        """ """
        return np.nanmedian( self.get_i_shift(use=use, **kwargs).data)

    def _get_delta_i_(self, which="sodium", method="mean", as_slice=True):
        """ """
        mu_nad_eff, mu_nad_eff_err = self.get_line_position(which, method=method)
        ijs_nad_eff = self.mapper.get_lbda_ij(mu_nad_eff)
        ijs_nad_ref = self.mapper.get_lbda_ij(self.reference[which])
        data_ = [ijs_nad_eff[traceindex][0]-ijs_nad_ref[traceindex][0]
                    for traceindex in self.cube.indexes]
        
        if not as_slice:
            return {traceindex:d_ for traceindex,d_ in zip(self.cube.indexes, data_)}
        
        return pyifu.get_slice(data_, self.cube.index_to_xy(self.cube.indexes), 
                                spaxel_vertices=self.cube.spaxel_vertices, 
                                indexes=self.cube.indexes)

        
    def get_i_shift(self, as_slice=True,
                        use=["sodium", "telluric"],
                        sodium_weight=5, method="mean"):
        """
        sodium_weight: [float] -optional-
            How much more weight to you wqnt to give to sodium in comparison to telluric
            = This is only used only if use=['sodium', 'telluric'] =
        """
        # LINE TO USE
        use = np.atleast_1d(use)
        if len(use) == 1:
            delta_i = self._get_delta_i_(which=use[0], method=method, as_slice=True)
            #delta_i.scale_by(-1) # To correct for it, remove it.
            return delta_i.data if as_slice else {i:d for i,d in zip(delta_i.indexes, delta_i.data)}

        elif list(use) not in [["sodium","telluric"],["telluric","sodium"]]:
            raise ValueError("use should either be 'sodium' or 'telluric' or both ['sodium', 'telluric']")

        sl_na   = self._get_delta_i_("sodium")
        sl_te   = self._get_delta_i_("telluric")
        data_   = np.average([sl_na.data,sl_te.data], weights=[sodium_weight, 1], axis=0) * -1
        sl      = pyifu.get_slice(data_, self.cube.index_to_xy(self.cube.indexes), 
                                spaxel_vertices=self.cube.spaxel_vertices, 
                                indexes=self.cube.indexes)

        #sl.scale_by(-1) # To correct for it, remove it.
        if not as_slice:
            return {i:d for i,d in zip(self.cube.indexes, sl.data)}

        return sl
            
            
    # -------- #
    #  LOADER  #
    # -------- #
    def load_sodium_spec(self, nspaxels=20, averaging=40, use_faintest=True):
        """ """
        if use_faintest:
            index_to_fit = np.asarray(self.cube.get_faintest_spaxels(nspaxels*averaging))
            np.random.shuffle(index_to_fit)
        else:
            index_to_fit= np.random.choice(np.arange(len(self.cube.indexes)), int(nspaxels*averaging), replace=False)

        indexes = index_to_fit.reshape(nspaxels,averaging)
        self._side_properties["spec_sodium_indexes"] = index_to_fit
        self._side_properties["spec_sodium"] = [self.cube.get_spectrum(index_,"data")
                                                   for index_ in indexes]
        
    def load_telluric_spec(self, nspaxels=5, averaging=50, nskyspaxels=100):
        """ """
        index_to_fit = np.asarray(self.cube.get_brightest_spaxels(nspaxels*averaging))
        np.random.shuffle(index_to_fit)
        indexes = index_to_fit.reshape(nspaxels,averaging)
        skyspec = self.cube.get_spectrum(self.cube.get_faintest_spaxels(nskyspaxels))
        spec_b = [self.cube.get_spectrum(index_,"data") for index_ in indexes]
        self._side_properties["spec_telluric"] = \
            [pyifu.get_spectrum(skyspec.lbda, spec_b_.data-skyspec.data, variance=spec_b_.variance)
                                        for spec_b_ in spec_b]
        
    def load_telluric_fit(self, wavebuffer=400):
        """ """
        if self.spec_telluric is None:
            self.load_telluric_spec()
        
        flaglocal = (self.cube.lbda > TELLURIC_REF_LBDA-wavebuffer) & (self.cube.lbda < TELLURIC_REF_LBDA+wavebuffer)
        fit_tell = []
        for spec_ in self.spec_telluric:
            fg = modefit.get_normpolyfit(spec_.lbda[flaglocal], spec_.data[flaglocal], spec_.data[flaglocal]/30, 2, 1)
            fg.fit(sig0_guess=30, sig0_boundaries=[2,None],
                       mu0_guess= self.reference["telluric"],
                       mu0_boundaries= [self.reference["telluric"]-30, self.reference["telluric"]+30])
            fit_tell.append(fg)
            
        self._derived_properties["fit_telluric"] = fit_tell
    
    def load_sodium_fit(self, wavebuffer=300):
        """ """
        if self.spec_sodium is None:
            self.load_sodium_spec()
        
        flaglocal = (self.cube.lbda > SODIUM_SKYLINE_LBDA-wavebuffer) & (self.cube.lbda < SODIUM_SKYLINE_LBDA+wavebuffer)
        fit_nad = []
        for spec_ in self.spec_sodium:
            fg = modefit.get_normpolyfit(spec_.lbda[flaglocal], spec_.data[flaglocal], spec_.data[flaglocal]/30, 2, 1)
            fg.fit(sig0_guess=10, sig0_boundaries=[2,None],
                       mu0_guess= self.reference["sodium"],
                       mu0_boundaries= [self.reference["sodium"]-50, self.reference["sodium"]+50])
            fit_nad.append(fg)
            
        self._derived_properties["fit_sodium"] = fit_nad
        
    # --------- #
    #  SETTER   #
    # --------- #
    def set_cube(self, cube_):
        """ attach to this intance a cube """
        self._properties["cube"] = cube_
        
    def set_mapper(self, mapper_):
        """ attach to this intance a mapper object containing the 
        ccd to cube coordinates conversions """
        self._properties["mapper"] = mapper_

    def set_reference(self, sodium=None, telluric=None):
        """ """
        if sodium is not None:
            self.reference["sodium"] = sodium
            
        if telluric is not None:
            self.reference["telluric"] = telluric

    # ================== #
    #   Properties       #
    # ================== #
    @property
    def cube(self):
        """ pyifu 3d cube """
        return self._properties["cube"]
    
    @property
    def mapper(self):
        """ object containing the ccd to 3d cube conversion tools """
        return self._properties["mapper"]

    @property
    def reference(self):
        """ object containing the ccd to 3d cube conversion tools """
        if self._properties["reference"] is None:
            self._properties["reference"] = {"sodium":SODIUM_SKYLINE_LBDA,
                                             "telluric":TELLURIC_REF_LBDA}
        return self._properties["reference"]
    
    #
    # Derived_property
    #
    @property
    def spec_sodium(self):
        """ """
        return self._side_properties["spec_sodium"]
    
    @property
    def spec_telluric(self):
        """ """
        return self._side_properties["spec_telluric"]
    
    @property
    def fit_sodium(self):
        """ """
        if self._derived_properties["fit_sodium"] is None:
            self.load_sodium_fit()
        return self._derived_properties["fit_sodium"]
    
    @property
    def fit_telluric(self):
        """ """
        if self._derived_properties["fit_telluric"] is None:
            self.load_telluric_fit()
        return self._derived_properties["fit_telluric"]
    
    @property
    def _sodium_indexes(self):
        """ """
        return self._side_properties["spec_sodium_indexes"]
    
###########################
#                         #
#  WaveSolution           #
#                         #
###########################
class WaveSolution( BaseObject ):
    """ """
    PROPERTIES = ["lamps"]
    DERIVED_PROPERTIES = ["wavesolutions","solutions"]

    # ================== #
    #  Main Methods      #
    # ================== #
    # -------- #
    #  I/O     #
    # -------- #
    def writeto(self, filename):
        """ save the object into the given filename (.pkl). 
        Load it the using the load() method.
        """
        from pysedm.utils.tools import dump_pkl
        dump_pkl(self.wavesolutions, filename)

    def load(self, filename):
        """ Load the object from the given filename (.pkl).
        object created by the writeto() method can be opened this way.
        """
        from pysedm.utils.tools import load_pkl
        data = load_pkl(filename)
        if "wavesolution" not in list(data.values())[0]:
            raise TypeError('The given dictionary file does not seem to be a wavelength solution. No "wavesolution" in the first entry')
        
        self.set_wavesolutions(data)
        
    # -------- #
    # BUILDER  #
    # -------- #
    def fit_wavelesolution(self, traceindex, sequential=True, contdegree=4, wavedegree=5,
                            saveplot=None, show=False, plotprop={}):
        """ Fit the wavelength solution of the given trace.

        This matching is made in two steps:
        1) fit the central positions of the loaded arcspectrum emission lines.
           This fit is based on linear combination of gaussian lines + polynomial continuum.
        2) Fit a polynomial degree to modelize the relation between the 
           fitted central positions (c.f. step 1) with the corresponding wavelengths
           of these lines.

        Parameters
        ----------
        traceindex: [int]
            index of the trace for which you want to fit the wavelength solution
            
        sequential: [bool] -optional-
            How the gaussian line profiles (central positions) should be fitted:
            - False: All at once (as if it was a unique spectrum)
            - True:  Once at the time (each arcspectrum fitted independently)

        contdegree: [int] -optional-
            The degree of the polynom underlying the gaussian lines
            
        wavedegree: [int] -optional-
            The degree of the polynom modelizing the pixel<->wavelength relation
            
        saveplot: [None or string] -optional-
            Save the wavelength solution figure. If None, nothing will be saved.
            
        show: [bool] -optional-
            Should the wavelength solution figure be shown?
        
        Returnsemus
        -------
        None
        """
         # <0.1s # with saved masks
        wsol_ = get_arccollection(traceindex, [self.lampccds[i] for i in self.lampnames])
        # 3s 
        wsol_.fit_lineposition(contdegree=contdegree, sequential=sequential)
        # 0.01s
        wsol_.fit_wavelengthsolution(wavedegree, legendre=False)
        self.wavesolutions[traceindex] = wsol_.data

                                         
        self._wsol = wsol_
        if saveplot is not None or show:
            wsol_.show(traceindex=traceindex, xrange=[3600,9500], savefile=saveplot, **plotprop)

    # -------- #
    # GETTER   #
    # -------- #
    def get_spaxel_wavesolution(self, traceindex):
        """ """
        if traceindex not in self.wavesolutions:
            raise ValueError("Unknown wavelength solution for the spaxels #%d"%traceindex)
        
        if traceindex not in self._solution:
            self._solution[traceindex] = SpaxelWaveSolution( self.wavesolutions[traceindex]["wavesolution"] )
            
        return self._solution[traceindex]

    def _load_full_solutions_(self):
        """ """
        for traceindex in self.wavesolutions.keys():
            if traceindex not in self._solution:
                self._solution[traceindex] = SpaxelWaveSolution(self.wavesolutions[traceindex]["wavesolution"])
                
            self._solution[traceindex].load_pixel_to_lbda_solution()
            
    
    def pixels_to_lbda(self, pixel, traceindex):
        """ Pick the requested spaxel and get the wavelength [in angstrom] that goes with the given pixel """
        return self.get_spaxel_wavesolution(traceindex).pixels_to_lbda(pixel)
    
    def lbda_to_pixels(self, lbda, traceindex):
        """ Pick the requested spaxel and get the pixel that goes with the given wavelength [in angstrom] """
        return self.get_spaxel_wavesolution(traceindex).lbda_to_pixels(lbda)
    
    # -------- #
    #  I/O     #
    # -------- #
    def set_wavesolutions(self, wavesolutions):
        """ Load a dictionary containing the wavelength solution of individual spaxels """
        [self.add_trace_wavesolution(traceindex, data) for traceindex,data in wavesolutions.items()]
        
            
    def add_trace_wavesolution(self, traceindex, data, replace=False):
        """ add to the current instance 'data' of a new wavelength solution.
        see SpaxelWaveSolution.data
        """
        if traceindex in self.wavesolutions.keys() and not replace:
            raise ValueError("%d is already loaded. Set `replace=True` to replace the current value."%traceindex)
        
        self.wavesolutions[traceindex] = data
        self._solution[traceindex]     = SpaxelWaveSolution(data["wavesolution"], datafitted=[data["usedlines"], data["fit_linepos"], data['fit_linepos.err']])
        

        
    def add_lampccd(self, lampccd, name=None):
        """ """
        if CCD not in lampccd.__class__.__mro__:
            raise TypeError("The given lampccd is not a pysedm's CCD object")
        
        if name is None:
            if lampccd.objname is None:
                raise ValueError("You need to provide a name for the arcspectrum")
            name = lampccd.objname
            
        # - Set it.
        self.lampccds[name] = lampccd

    def remove_lampccd(self, name):
        """ """
        self.lampccds.pop(name)

    # --------- #
    # PLOTTER   #
    # --------- #
    def show_dispersion_map(self, hexagrid,
                                ax=None, savefile=None, show=True,
                                kind="nMAD", vmin="0.5", vmax="99.5",
                            clabel=r"nMAD [$\AA$]", **kwargs):
        """ """
        from pysedm.sedm import display_on_hexagrid
        traceindexes = list(self._solution.keys())
        value = nmad = [self._solution[i].get_wavesolution_rms(kind="nMAD") for i in traceindexes]
        return display_on_hexagrid(value, traceindexes,hexagrid=hexagrid, 
                                       ax=ax, vmin=vmin, vmax=vmax,
                                       clabel=clabel, savefile=savefile, show=show,
                                       **kwargs)

    # ---------- #
    # STATISTICS #
    # ---------- #
    def calc_dispersion_stats(self, savefile=None):
        """ Calculate disperson solution statistics """

        traceindexes = list(self._solution.keys())

        rms = [self._solution[i].get_wavesolution_rms(kind="rms")
               for i in traceindexes]
        mad = [self._solution[i].get_wavesolution_rms(kind="nmad")
               for i in traceindexes]
        wrms = [self._solution[i].get_wavesolution_rms(kind="wrms")
                for i in traceindexes]
        if savefile:
            stat_f = open(savefile, "w")
            stat_f.write("NSpax: %d\n" % len(rms))
            stat_f.write("MinRMS: %.3f\n" % min(rms))
            stat_f.write("AvgRMS: %.3f\n" % np.nanmean(rms))
            stat_f.write("MaxRMS: %.3f\n" % max(rms))
            stat_f.write("MinMAD: %.3f\n" % min(mad))
            stat_f.write("AvgMAD: %.3f\n" % np.nanmean(mad))
            stat_f.write("MaxMAD: %.3f\n" % max(mad))
            stat_f.write("MinWRMS: %.3f\n" % min(wrms))
            stat_f.write("AvgWRMS: %.3f\n" % np.nanmean(wrms))
            stat_f.write("MaxWRMS: %.3f\n" % max(wrms))
            stat_f.close()
        print("nspax, min, avg, max RMS: %d, %.3f, %.3f, %.3f" %
              (len(rms),  min(rms), np.nanmean(rms), max(rms)))
        return len(rms), min(rms), np.nanmean(rms), max(rms)

    # ================== #
    #  Properties        #
    # ================== #
    @property
    def lampnames(self):
        """ Names of the lamps loaded """
        return np.sort(list(self.lampccds.keys()))
    @property
    def lampccds(self):
        """ dictionary containing the ScienceCCD objects used
        to get the wavelength solution of the cube """
        if self._properties["lamps"] is None:
            self._properties["lamps"] = {}
        return self._properties["lamps"]

    @property
    def wavesolutions(self):
        """ dictionary containing the wavelength solution for the loaded spectra """
        if self._derived_properties["wavesolutions"] is None:
            self._derived_properties["wavesolutions"] = {}
        return self._derived_properties["wavesolutions"]
    @property
    def _solution(self):
        """ WaveSolution object. use get_wavesolution() """
        if self._derived_properties["solutions"] is None:
            self._derived_properties["solutions"] = {}
        return self._derived_properties["solutions"]

    @property
    def traceindexes(self):
        """ list of indexes having wavelengh solution loaded. """
        return list(self.wavesolutions.keys())
    
class SpaxelWaveSolution( BaseObject ):
    """ """
    PROPERTIES = ["wavesolution","inverse_wavesolution"]
    SIDE_PROPERTIES = ["datafitted"]
    def __init__(self, polycoefs=None, datafitted=None):
        """ 
        Parameters: 
        -----------
        polycoef: [array] -optional-
            The polynomial's coefficients, in decreasing powers.
            For example, ``[1, 2, 3]`` returns an object that represents
            :math:`x^2 + 2x + 3`

        datafitted: [3 arrays] -optional-
            You could provide x,y and dy.
            The data used to build the SpaxelWaveSolution

        Returns
        -------
        """
        self.__build__()
        if polycoefs is not None:
            self.set_solution(polycoefs)
        if datafitted is not None:
            self.set_datafitted(*datafitted)
            
    # ================== #
    #  Main Methods      #
    # ================== #
    # ---------- #
    #  GETTER    #
    # ---------- #
    def get_wavesolution_rms(self, kind="wrms"):
        """ """
        if self.datafitted is None:
            raise AttributeError("This needs the self.datafitted to work")

        res = self.pixels_to_lbda(self.datafitted['fit_linepos'])-self.datafitted['usedlines']
        known_kind = ["std or rms", "nmad", "wrms"]
        
        if kind in ["std", "rms"]:
            return np.nanstd(res)
        if kind in ['nMAD', 'nmad','mad_std']:
            from astropy.stats import mad_std
            return mad_std(res)
        if kind in ["wrms", "wRMS"]:
            err   = self.datafitted['fit_linepos.err']
            wmean = np.average(np.average(res, weights=1./err**2))
            return np.sqrt(np.average((res-wmean)**2, weights=1./err*2))
        raise ValueError("unknown kind %s."%kind+" use: "+", ".join(known_kind))
    # ---------- #
    #  SETTER    #
    # ---------- #
    def set_solution(self, polycoef ):
        """ Sets the wavelength solution based on numpy's poly1d
        
        Parameters: 
        -----------
        polycoef: [array]
            The polynomial's coefficients, in decreasing powers.
            For example, ``[1, 2, 3]`` returns an object that represents
            :math:`x^2 + 2x + 3`

        Returns
        -------
        Void 
        """
        self._properties["wavesolution"] = np.poly1d(polycoef)

    def set_datafitted(self, usedlines, fit_linepos, fit_lineposerr):
        """ The expected wavelength, the fitted line position in pixels and its associated error """
        self._side_properties['datafitted'] = {"usedlines":usedlines,
                                                "fit_linepos":fit_linepos,
                                                "fit_linepos.err":fit_lineposerr}
        
    def writeto(self, filename, **kwargs):
        """ dump the `data` into the given `filename` with the given `format`.
        - This uses numpy.savetxt -
        """
        np.savetxt(filename, self.data, **kwargs)

    def load(self, filename, **kwargs):
        """ Loads a file containing the polynomial coefficients.
        The e.g. reads files created by `writeto`.
        - This uses numpy loadtxt -
        """
        self.set_solution( np.loadtxt(filename, **kwargs))

    # -------- #
    # PLOTTER  #
    # -------- #
    def show(self, ax=None, savefile=None, show=True, ecolor="0.3",
                 xrange=None, **kwargs):
        """ """
        from astrobject.utils.mpladdon import figout, errorscatter
        from astrobject.utils.tools    import kwargs_update
        
        if self.datafitted is None:
            raise AttributeError('No datafitted to be shown')
        
        # - Non Sequential
        if ax is None:
            fig = mpl.figure(figsize=[6,5])
            ax = fig.add_axes([0.12,0.12,0.78,0.78])
            ax.set_xlabel(r"Wavelength [$\mathrm{\AA}$]", fontsize="large")
            ax.set_ylabel(r"Pixels (ccd-row)", fontsize="large")
        else:
            fig = ax.figure

        prop = kwargs_update( dict(mfc=mpl.cm.binary(0.2,0.5), mec=mpl.cm.binary(0.8,0.5),
                                       ms=15, ls="None",mew=1.5, marker="o", zorder=5), **kwargs)
        ax.plot(self.datafitted["usedlines"], self.datafitted["fit_linepos"], **prop)
        er = ax.errorscatter(self.datafitted["usedlines"], self.datafitted["fit_linepos"],
                                 dy=self.datafitted["fit_linepos.err"], zorder=prop["zorder"]-1,
                             ecolor=ecolor)

        if xrange is None:
            xrange = [self.datafitted["usedlines"].min()-100, self.datafitted["usedlines"].max()+100]
        x = np.linspace(xrange[0],xrange[1], 1000)
        ml = ax.plot(x, self.lbda_to_pixels(x), lw=2, color="C1")
        
        fig.figout(savefile=savefile, show=show)
        
    # ================== #
    #   Properties       #
    # ================== #
    @property
    def _wavesolution(self):
        """ """
        return self._properties["wavesolution"]
    
    @property
    def data(self):
        """ Coefficents of the poly1d parameters.
        do set_solution(self.data) 
        """
        if not self.has_wavesolution():
            return None
        return self._wavesolution.coeffs
    
    @property
    def datafitted(self):
        """ """
        return self._side_properties["datafitted"]
    
    # - SpaxelWaveSolution
    def has_wavesolution(self):
        """ Test if the wavelength solution has been set. """
        return self._wavesolution is not None
    
    
    def lbda_to_pixels(self, wavelength):
        """ numpy.poly1d by the best fitted wavelength solution fit.

        Parameters
        ----------
        wavelength: [float or array]
            Wavelength in Angstrom

        Returns
        -------
        pixel (or array of)

        Example:
        --------
        self.wavesolution(6500) provides the pixel 
        corresponding to the 6500 Angstrom wavelength
        """
        return self._wavesolution(wavelength-REFWAVELENGTH)

    def load_pixel_to_lbda_solution(self):
        """ """
        if _HASPYNVERSE:
            self._properties["inverse_wavesolution"] = inversefunc(self.lbda_to_pixels)
        else:
            raise ImportError("You need pynverse | pip install pynverse")
            
    @property
    def pixels_to_lbda(self):
        """ Convert Pixels to Lambda 

        Example:
        --------
        self.inverse_wavesolution(1880) provides the wavelength (Angstrom)
        corresponding to the 1880th pixel
        """
        if self._properties["inverse_wavesolution"] is None:
            self.load_pixel_to_lbda_solution()

        return self._properties["inverse_wavesolution"]

###########################
#                         #
#  Arc Spectrum           #
#                         #
###########################
class VirtualArcSpectrum( BaseObject ):
    """ Virtual Class Upon Wich ArcSpectrum and ArcSpectrumCollection is built """

    PROPERTIES         = ["wavesolution"]
    DERIVED_PROPERTIES = ["linefitter", "solutionfitter"]
    
    # ================ #
    #  Main Method     #
    # ================ #
    # --------- #
    #  GETTER   #
    # --------- #
    def get_line_shift(self):
        """ Shift of the central line value based on the considered spaxel """
        if self.arcname in ["Hg"]:
            wavemax = np.max(self.lbda[self.get_arg_maxflux(2)])
        elif self.arcname in ["Xe"]:
            wavemax = np.min(self.lbda[self.get_arg_maxflux(2)])
        else:
            wavemax = self.lbda[self.get_arg_maxflux(1)]
          
          
        wavemax_expected = self.arclines[self.expected_brightesline]["mu"]
        return wavemax-wavemax_expected
    
    def get_arg_maxflux(self, nbest, mask=None, order=3, **kwargs):
        """ The argument of the maximum values. 

        Parameters
        ----------
        nbest: [int]
            Number of local maximum you want. 
            
        mask: [boolean array] -optional-
            Where to look. (value True in mask are kept)
            
        order: [int] -optional-
            How many points on each side to use for the comparison
            to consider ``comparator(n, n+x)`` to be True.

        **kwargs goes to scipy.signal.argrelextrema()
        
        Returns
        -------
        nd-array (n=nbest)
        """
        from scipy.signal import argrelextrema
        f = self.flux if mask is None else self.flux[mask]
        args = argrelextrema(f, np.greater,order=order, **kwargs)[0]
        return args[np.argsort(f[args])][-nbest:]

    # --------- #
    #  FITTER   #
    # --------- #
    def fit_wavelengthsolution(self, wavesolution_degree=3, legendre=False,**kwargs):
        """ fit the 'pixel<->wavelength' solution 
        
        Parameters
        ----------
        wavesolution_degree: [degree] -optional-
            degree of the polynom used to fit the 'pixel<->wavelength' solution

        legendre: [bool] -optional-
            should the fitted polynom be based on Legendre polynoms?

        **kwargs goes to fit_lineposition()
            
            contdegree: [int] -optional-
                Degree of the (Legendre) polynom used as continuum

            line_shift: [float] -optional-
                Force the expected line pixel shift used as first guesses. 
                If not provided an internal procedure based on expected and observed 
                brightest emission position will be used. 
                (see `get_line_shift()`)
        
            exclude_reddest_part: [bool] -optional-
                If True, wavelengths that are more than *red_buffer* [wave units] redder
                than the reddest 'usedline' will be excluded from the fit.
                If activated, this option raises a warning.

            red_buffer: [float] -optional-
                How much redder than the reddest emission line should the fit conserve.
                This is ignored if *exclude_reddest_part* is False

        Returns
        -------
        Void
        """
        from modefit import get_polyfit
        
        if not self.has_linefitter():
            self.fit_lineposition(**kwargs)
            
        mus, emus = self._linefit_to_mus_()
        self._derived_properties["solutionfitter"] = get_polyfit( self.usedlines-REFWAVELENGTH, mus, emus+0.2,
                                                                      wavesolution_degree,
                                                                      legendre=legendre)
        
        guesses = {"a0_guess":np.nanmean(mus)}
        if not legendre:
            if wavesolution_degree>=2:
                guesses["a1_guess"]= 0.05
                
        self.guesses = guesses.copy()
        self.solutionfitter.fit(**guesses)
        self.datafitted = [self.usedlines, mus, emus]
        # - Set the best fit solution
        self._derived_properties["wavesolution"] = \
          SpaxelWaveSolution([self.solutionfitter.fitvalues["a%i"%i]
                        for i in range(self.solutionfitter.model.DEGREE)[::-1]],
                            datafitted=self.datafitted)

    def _load_lineposition_(self, contdegree=2, line_shift=None,
                             exclude_reddest_part=True,
                             red_buffer=30,
                             exclude_bluest_part=True,
                             blue_buffer=30, line_to_skip=None
                              ):
        """ Fit gaussian profiles of expected arclamp emmisions.
        The list of fitted lines are given in `usedlines`.
        The fitter object is stored as `linefitter`.
        
        - This method uses `modefit` -

        Pamameters
        ----------
        contdegree: [int] -optional-
            Degree of the (Legendre) polynom used as continuum

        line_shift: [float] -optional-
            Force the expected line pixel shift used as first guesses. 
            If not provided an internal procedure based on expected and observed 
            brightest emission position will be used. 
            (see `get_line_shift()`)
        
        exclude_reddest_part: [bool] -optional-
            If True, wavelengths that are more than *red_buffer* [wave units] redder
            than the reddest 'usedline' will be excluded from the fit.
            If activated, this option raises a warning.

        red_buffer: [float] -optional-
           How much redder than the reddest emission line should the fit conserve.
           This is ignored if *exclude_reddest_part* is False

        Returns
        -------
        Void (sets linefitter)
        """
        from modefit import get_normpolyfit
        
        # where to look at? (~1ms)
        flagin = (self.lbda>=self.databounds[0])  * (self.lbda<=self.databounds[1]) # 1ms

        # Building guess (~1ms)
        self._normguesses = {}
        if line_shift is None:
            lines_shift = self.get_line_shift()
            
        for i,l in enumerate(self.usedlines):
            self._normguesses["ampl%d_guess"%i]      = self.arclines[l]["ampl"]
            self._normguesses["ampl%d_boundaries"%i] = [self.arclines[l]["ampl"]*0.2, self.arclines[l]["ampl"]*3]
            
            self._normguesses["mu%d_guess"%i]        = self.arclines[l]["mu"]+lines_shift
            self._normguesses["mu%d_boundaries"%i]   = [self._normguesses["mu%d_guess"%i]-2, self._normguesses["mu%d_guess"%i]+2]
            
            self._normguesses["sig%d_guess"%i]       = 1.1 if (not "doublet" in self.arclines[l] or not self.arclines[l]["doublet"]) else 1.8
            self._normguesses["sig%d_boundaries"%i]  = [0.9,1.5] if (not "doublet" in self.arclines[l] or not self.arclines[l]["doublet"]) else [1.1, 3]

        # where do you wanna fit? (~1ms)
        if exclude_reddest_part:
            flagin *= (self.lbda<=self._normguesses["mu%d_guess"%(len(self.usedlines)-1)]+red_buffer)
        else:
            warnings.warn("part redder than %d *not* removed"%(self._normguesses["mu%d_guess"%(len(self.usedlines)-1)]+red_buffer))
            
        if exclude_bluest_part:
            flagin *= (self.lbda>=self._normguesses["mu0_guess"]-blue_buffer)
        else:
            warnings.warn("part bluer than %d *not* removed"%(self._normguesses["mu0_guess"]-blue_buffer))

        # Setup the linefitter (3ms)
        norm = np.nanmean(self.flux[flagin])
        lbdas = self.lbda[flagin].copy()
        errors = self.errors[flagin]/norm if self.has_errors() else np.nanstd(self.flux[flagin])/norm/5.
        
        self._derived_properties["linefitter"] = \
          get_normpolyfit(lbdas,self.flux[flagin]/norm,
                              errors,
                              contdegree, ngauss=len(self.usedlines), legendre=True)

    def fit_lineposition(self, contdegree=2, line_shift=None,
                             exclude_reddest_part=True,
                             red_buffer=30,
                             exclude_bluest_part=True,
                             blue_buffer=30, line_to_skip=None
                             ):
        # VirtualArcSpectrum
        """ Fit gaussian profiles of expected arclamp emmisions.
        The list of fitted lines are given in `usedlines`.
        The fitter object is stored as `linefitter`.
        
        - This method uses `modefit` -

        Pamameters
        ----------
        contdegree: [int] -optional-
            Degree of the (Legendre) polynom used as continuum

        line_shift: [float] -optional-
            Force the expected line pixel shift used as first guesses. 
            If not provided an internal procedure based on expected and observed 
            brightest emission position will be used. 
            (see `get_line_shift()`)
        
        exclude_reddest_part: [bool] -optional-
            If True, wavelengths that are more than *red_buffer* [wave units] redder
            than the reddest 'usedline' will be excluded from the fit.
            If activated, this option raises a warning.

        red_buffer: [float] -optional-
           How much redder than the reddest emission line should the fit conserve.
           This is ignored if *exclude_reddest_part* is False

        Returns
        -------
        Void (sets linefitter)
        """
        
        if not self.has_linefitter():
            self._load_lineposition_(contdegree=contdegree,
                                         line_shift=line_shift,
                                         exclude_reddest_part=exclude_reddest_part,
                                         red_buffer=red_buffer,
                                         exclude_bluest_part=exclude_bluest_part,
                                         blue_buffer=blue_buffer, line_to_skip=line_to_skip)
        # The actual fit ~4s
        self._normguesses["a0_guess"] = np.percentile(self.linefitter.data, 25)

        self.linefitter.fit( **self._normguesses )

    def _linefit_to_mus_(self):
        """ returns the best fit values ordered as the usedlines 
        Internal tools to go from linefitter -> wavefitter
        """
        return np.asarray([[self.linefitter.fitvalues["mu%d"%i],
                            self.linefitter.fitvalues["mu%d.err"%i]]
                           for i,l in enumerate(self.usedlines)]).T

    # ------------------ #
    #  Apply Solution    #
    # ------------------ #        
    def pixels_to_lbda(self, pixels):
        """ """
        if self.has_wavesolution():
            return self.wavesolution.pixels_to_lbda(pixels)
        
        raise AttributeError("No wavelength solution define. Run fit_wavesolution()")
            
    def lbda_to_pixels(self, lbda):
        """ based on the wavelength solution loaded (see 'wavesolution'), 
        this converts an lbda in angstrom to a pixel 

        Return
        ------
        array (or float, see lbda)
        """
        if self.has_wavesolution():
            return self.wavesolution.lbda_to_pixels(lbda)
        
        raise AttributeError("No wavelength solution define. Run fit_wavesolution()")
    
    # ================ #
    #  Properties      #
    # ================ #
    @property
    def data(self):
        """ Foundamental Information about the fit """
        mus,emus = self._linefit_to_mus_() if self.has_linefitter() else [None,None]
        if self.has_linefitter():
            fitvalue = self.linefitter.fitvalues if not self._sequentialfit \
              else {lamp:self.linefitter[lamp].fitvalues for lamp in self.arcnames}
        else:
            fitvalue = None
        return {"lampname": self.arcnames,
                "usedlines": self.usedlines,
                "fit_linepos":mus,"fit_linepos.err":emus,
                "wavesolution": self.wavesolution.data if self.has_wavesolution() else None,
                "line_fitvalues":fitvalue
                    }

    def set_databounds(self, xmin, xmax):
        """ """
        self._properties["databounds"] = [xmin, xmax]
        
    @property
    def databounds(self):
        """ limits of the effective spectrum """
        return self._properties["databounds"]

    # -----------
    # - Lines
    @property
    def usedlines(self):
        """ Wavelengthes used for the wavelength matching """
        return np.sort(list(self.arclines.keys()))
    
    @property
    def expected_brightesline(self):
        """ lines that has the greatest amplitude in the LINES global variable """
        l, amp = np.asarray([[l,v["ampl"]] for l,v in self.arclines.items()]).T
        return l[np.argmax(amp)]

    # - SpaxelWaveSolution
    @property
    def wavesolution(self):
        """ """
        return self._derived_properties['wavesolution']
    
    def has_wavesolution(self):
        """ """
        return self.wavesolution is not None
    
    # -------------
    #  Fitters
    # - Lines
    @property
    def linefitter(self):
        """ Object used to fit the gaussian lines. 
        The fitted central values are then used for the solutionfitter() """
        return self._derived_properties["linefitter"]

    def has_linefitter(self):
        """ Has the Gaussian line fitter has been set? """
        return self.linefitter is not None
    
    # - Wavelength Solution    
    @property
    def solutionfitter(self):
        """ Object that enables to fit the wavelength solution """
        return self._derived_properties["solutionfitter"]
    
    def has_solutionfitter(self):
        """ Has the wavelength solution fitter has been set?"""
        return self.solutionfitter is not None
    
class ArcSpectrum( Spectrum, VirtualArcSpectrum ):
    """ Arc lamp Spectrum object containing line fitting and wavelength solution fitting """
    
    PROPERTIES = ["name"]
    
    DERIVED_PROPERTIES = ["arclines"]
            
    # ================ #
    #  Main Method     #
    # ================ #
    # --------- #
    #  GETTER   #
    # --------- #
    def line_to_fitnumber(self, line):
        """ What is the index of the line within the 'usedlines' array """
        return np.argwhere(self.usedlines==line)[0]
            # --------- #
    #  SETTER   #
    # --------- #
    def set_arcname(self, name):
        """ Provide the name of the arclamp. 
        
        Parameters
        ----------
        name: [string]
            The given arcname should be in the list of predefined LINES
            otherwise you will not be able to acess 
            - usedlines
            - arclines
        
        Returns
        -------
        """
        self._properties["name"] = name
        if name not in LINES.keys():
            warnings.warn("Unknown Arc Lamp (%s). No Line Attached"%name)
            self._derived_properties["arclines"] = {}
        else:
            self.reload_lines()

    def remove_line(self, line):
        """ """
        self.arclines.pop(line)

    def reload_lines(self):
        """ """
        self._derived_properties["arclines"] = LINES[self.arcname].copy()
        
    # ================ #
    #  Properties      #
    # ================ #
    @property
    def arcname(self):
        """ Name of arc lamp"""
        return self._properties["name"]
    
    @property
    def arclines(self):
        """ ArcLines dictionary containing expected amplitude and central values """
        if self._derived_properties["arclines"] is None:
            self._derived_properties["arclines"] = {}
        return self._derived_properties["arclines"]

    @property
    def flux(self):
        """ """
        return self.data
    
    @property
    def errors(self):
        """ """
        return np.sqrt(self.variance) if self.has_variance() else None
    
    def has_errors(self):
        """ """
        return self.errors is not None

###########################
#                         #
#  Collection of Arcs     #
#                         #
###########################
class ArcSpectrumCollection( VirtualArcSpectrum ):
    """  """
    PROPERTIES = ["arcspectra"]
    DERIVED_PROPERTIES = ["arclines", "lbda"]
    
    # ================ #
    #  Main Methods    #
    # ================ #
    def get_line_shift(self):
        """ Shift of the central line value based on the considered spaxel.
        This is based on sequential check Cd then Hd then Xe. 
        First find, first pick.
        
        Returns
        -------
        float
        """
        
        if "Cd" in self.arcspectra:
            return self.arcspectra["Cd"].get_line_shift()
        if "Hg" in self.arcspectra:
            return self.arcspectra["Hg"].get_line_shift()
        if "Xe" in self.arcspectra:
            return self.arcspectra["Xe"].get_line_shift()
        raise AttributeError("known of the pre-defined lines (Cd, Hg and Xe) have been found in the arcspectra."+\
                              " They are requested for the line_shift ")

    # -------- #
    #  FITTER  #
    # -------- #
    def fit_lineposition(self, sequential=True,
                             contdegree=2, line_shift=None,
                             exclude_reddest_part=True,
                             red_buffer=30,
                             exclude_bluest_part=True,
                             blue_buffer=30
                             ):
        # ArcSpectrumCollection
        """ Fit gaussian profiles of expected arclamp emmisions.
        The list of fitted lines are given in `usedlines`.
        The fitter object is stored as `linefitter`.
        
        - This method uses `modefit` -

        Pamameters
        ----------
        sequential: [bool] -optional-
            Shall the fit be made line by line?

        contdegree: [int] -optional-
            Degree of the (Legendre) polynom used as continuum

        line_shift: [float] -optional-
            Force the expected shift of the lines that are used as first guesses 
            for the fit. If not provided, an internal automatic procedure based
            of the expected and observed brightest emission position is used.
            (see `get_line_shift()`)
        
        exclude_reddest_part: [bool] -optional-
            If True, wavelengths that are more than *red_buffer* [wave units] redder
            than the reddest 'usedline' will be excluded from the fit.
            If activated, this option raises a warning.

        red_buffer: [float] -optional-
           How much redder than the reddest emission line should the fit conserve.
           This is ignored if *exclude_reddest_part* is False

        Returns
        -------
        Void (sets linefitter)
        """
        self._build_arclines_()
        
        lineprop = dict(contdegree=contdegree, line_shift=line_shift,
                        exclude_reddest_part=exclude_reddest_part,
                        red_buffer=red_buffer,
                        exclude_bluest_part=exclude_bluest_part,
                        blue_buffer=blue_buffer)
        
        if not sequential:
            super(ArcSpectrumCollection, self).fit_lineposition(**lineprop )
            self._sequentialfit = False
        else:
            [self.arcspectra[s].fit_lineposition(**lineprop) for s in self.arcnames]
            self._derived_properties["linefitter"] = {s:self.arcspectra[s].linefitter for s in self.arcnames}
            self._sequentialfit = True

    def _linefit_to_mus_(self):
        """ returns the best fit values ordered as the usedlines 
        Internal tools to go from linefitter -> wavefitter
        """
        # - Non Sequential
        if not self.is_linefit_sequential():
            return super(ArcSpectrumCollection, self)._linefit_to_mus_()
        
        # - Sequential
        mus, emus = [],[]
        for li in self.usedlines:
            lamp_ = self.line_to_lamp(li)
            line_ = self.arcspectra[lamp_].line_to_fitnumber(li)[0]
            mus.append(self.linefitter[lamp_].fitvalues["mu%d"%line_])
            emus.append(self.linefitter[lamp_].fitvalues["mu%d.err"%line_])
            
        return np.asarray(mus),np.asarray(emus)
    
    # -------- #
    #  IO      #
    # -------- #
    def line_to_lamp(self, line):
        """ """
        return self.arclines[line]["arcname"]
    
    def add_arcspectrum(self, arcspectrum, name=None):
        """ """
        if ArcSpectrum not in arcspectrum.__class__.__mro__:
            raise TypeError("The given arcspectrum is not a pysedm's ArcSpectrum object")
        if name is None:
            if arcspectrum.arcname is None:
                raise ValueError("You need to provide a name for the arcspectrum")
            name = arcspectrum.arcname
            
        # - Set it.
        self.arcspectra[name] = arcspectrum
        self._derived_properties["arclines"] = None

    def remove_arcspectrum(self, arcname):
        """ Remove an arcspectrum from the collection """
        self.arcspectra.pop(arcname)
        self._derived_properties["arclines"] = None


    # -------- #
    # PLOTTER  #
    # -------- #
    def show(self, savefile=None, **kwargs):
        """ """
        if self.has_linefitter() and self.has_wavesolution():
            self._show_full_(savefile=savefile, **kwargs)
        elif self.has_wavesolution():
            self._show_wavesolution_(savefile=savefile, **kwargs)
        elif self.has_linefitter():
            self._show_linefit_(savefile=savefile, **kwargs)
        else:
            raise AttributeError("No SpaxelWaveSolution, Not LineFitter. Nothing to show")

    def _show_full_(self, savefile=None, show=True,
                    fig=None, stampsloc="right", traceindex=None,
                        show_guesses=False, **kwargs):
        """ """
        from astrobject.utils.mpladdon import figout
        if fig is None:
            fig = mpl.figure(figsize=[10,5]) if stampsloc in ["left", "right"] \
              else mpl.figure(figsize=[10,10])

        if stampsloc == "right":
            heigth, bottom = 0.82, 0.12
            space = 0.13
            axwave   = fig.add_axes([0.1,bottom,0.58,heigth])
            axstamps = [fig.add_axes([0.7,bottom+((heigth+space/(self.nspectra-1))/self.nspectra)*i,
                                      0.25, (heigth-space)/self.nspectra])
                            for i in range(self.nspectra)]
                
            self._show_linefit_(show=False, savefile=None, ax=axstamps, remove_xticks=False,
                                    show_guesses=show_guesses)
            self._show_wavesolution_(show=False, savefile=None, ax=axwave, **kwargs)

            axwave.set_xlabel(r"Wavelength [$\mathrm{\AA}$]", fontsize="large")
            axwave.set_ylabel(r"Pixels (ccd-row)", fontsize="large")
            axstamps[0].set_xlabel(r"Pixels (ccd-row)", fontsize="medium")
            if traceindex is not None:
                fig.text(0.01,0.99,"Spectrum #%d"%traceindex,
                        va="top",ha="left", fontsize="small")

        fig.figout(savefile=savefile, show=show)
        
    def _show_linefit_(self, savefile=None, show=True, ax=None,
                           show_gaussian=True, show_guesses=False,
                           remove_xticks=False, **kwargs):
        """ """
        from astrobject.utils.mpladdon import figout
        if not self._sequentialfit:
            # - Non Sequential
            if ax is None:
                fig = mpl.figure(figsize=[8,5])
                ax = fig.add_axes([0.12,0.12,0.78,0.78])
                ax.set_xlabel(r"Wavelength [$\mathrm{\AA}$]", fontsize="large")
                ax.set_ylabel(r"Flux []", fontsize="large")
            else:
                fig = ax.figure

            self.linefitter.show(ax=ax, savefile=None, show=False, show_gaussian=show_gaussian, **kwargs)
            if remove_xticks:
                ax.set_xticks([])
            if show_guesses:
                    _ = [ax.axvline(self.linefitter[lamp].param_input["mu%s_guess"%i], color="0.8", lw=1) 
                         for i,l in enumerate(self.usedlines)]
        else:
            # - Sequential
            if ax is None:
                fig = mpl.figure(figsize=[10,3])
                ax = [fig.add_subplot(1,self.nspectra,i+1) for i in range(self.nspectra)]
            elif len(ax) != self.nspectra:
                raise ValueError("In Sequential fit, ax must be an array with the size of `nspectra`")
            else:
                fig = ax[0].figure
                
            for ax_, lamp in zip(ax, self.arcnames):
                self.linefitter[lamp].show(ax=ax_, savefile=None, show=False,
                                        show_gaussian=show_gaussian,**kwargs)
                color = self._lamp_to_color_(lamp)
                [[s_.set_color(color),s_.set_linewidth(1.)] for s_ in ax_.spines.values()]
                ax_.text(0.04,0.96, lamp, transform=ax_.transAxes,
                         va="top", ha="left", fontsize="medium")
                ax_.set_yticks([])
                if remove_xticks:
                    ax_.set_xticks([])
                if show_guesses:
                    _ = [ax_.axvline(self.linefitter[lamp].param_input["mu%s_guess"%i], color="0.8", lw=1) 
                         for i,l in enumerate(self.arcspectra[lamp].usedlines)]

                
        fig.figout(savefile=savefile, show=show)
            
    def _show_wavesolution_(self, savefile=None, show=True, ax=None,
                                show_legend=True, show_model=True,
                                ecolor="0.3", xrange=None,**kwargs):
        """ """
        from astrobject.utils.mpladdon import figout, errorscatter
        from astrobject.utils.tools    import kwargs_update
        
        self._plot = {}
        if ax is None:
            fig = mpl.figure(figsize=[8,5])
            ax = fig.add_axes([0.12,0.12,0.78,0.78])
            ax.set_xlabel(r"Wavelength [$\mathrm{\AA}$]", fontsize="large")
            ax.set_ylabel(r"Pixels (ccd-row)", fontsize="large")
        else:
            fig = ax.figure
            
        # -------
        # - Data
        prop = kwargs_update( dict(ms=15, ls="None",mew=1.5, marker="o", zorder=5), **kwargs)
        mu,emus = self._linefit_to_mus_()
        # Fancy coloring based on lamps
        for i,lamp in enumerate(self.arcnames[::-1]):
            flagin = np.asarray([lamp in self.line_to_lamp(l)  for l in self.usedlines], dtype="bool")
            ax.plot(self.usedlines[flagin], mu[flagin],
                        mfc= self._lamp_to_color_(lamp, 0.5),mec= self._lamp_to_color_(lamp, 0.9),
                        label=r"%s"%lamp,**prop)

        er = ax.errorscatter(self.usedlines, mu, dy=emus, zorder=prop["zorder"]-1,
                             ecolor=ecolor)
        # --------
        # - Model
        if show_model:
            if xrange is None:
                xrange = [self.usedlines.min()-100, self.usedlines.max()+100]
            x = np.linspace(xrange[0],xrange[1], 1000)
            ml = ax.plot(x, self.lbda_to_pixels(x), lw=2, color="C1")
        # --------
        # - Output
        if show_legend:
            ax.legend(loc="upper left", frameon=False, fontsize="medium",
                          markerscale=0.6)

        fig.figout(savefile=savefile, show=show)

    def _lamp_to_color_(self, lamp, alpha=1):
        """ Internal Ploting tool to get a consistent color for the lamp.
        Slower to use that, but prettier."""
        from pysedm.utils.mpl import get_lamp_color
        return get_lamp_color(lamp, alpha=alpha)
        
    # ================ #
    #  Properties      #
    # ================ #
    # -----------
    # Fake Spectrum
    @property
    def lbda(self):
        """ """
        if self._derived_properties['lbda'] is None and self.has_spectra():
            self._derived_properties['lbda'] = list(self.arcspectra.values())[0].lbda
        return self._derived_properties['lbda']
    
    @property
    def flux(self):
        return np.sum([s.flux for s in self.arcspectra.values()], axis=0)\
          if self.has_spectra() else None
    
    @property
    def errors(self):
        """ """
        #TO BE DONE
        return None
    
    def has_flux(self):
        return self.flux is not None
    
    def has_errors(self):
        return self.errors is not None

    # -----------
    # Collection
    @property
    def arcspectra(self):
        """ dictionary containing the individual arcspectrum set """
        if self._properties["arcspectra"] is None:
            self._properties["arcspectra"] = {}
        return self._properties["arcspectra"]
    
    @property
    def nspectra(self):
        """ number of arcspectrum attached to this collection """
        return len(self.arcspectra.keys())
    
    def has_spectra(self):
        """ Tests if you already loaded spectra"""
        return self.nspectra>0

    @property
    def arcnames(self):
        return np.sort(list(self.arcspectra.keys()))
    
    # -----------------
    # Collection Tricks
    def is_linefit_sequential(self):
        """ """
        if not self.has_linefitter():
            return None
        return self._sequentialfit
    
    @property
    def arclines(self):
        """ """
        if self._derived_properties["arclines"] is None:
            self._build_arclines_()
        return self._derived_properties["arclines"]
    

    def _build_arclines_(self, rebuild=False):
        """ """
        if self._derived_properties["arclines"] is not None and not rebuild:
            return
        di={}
        for lampname,d in list(self.arcspectra.items()):
            for k,v in list(d.arclines.items()):
                if "backup" in v.keys() and v["backup"] in self.arcnames:
                    warnings.warn("line %s skipped since %s loaded"%(k,v["backup"]))
                    d.remove_line(k)
                    continue
                v["arcname"] = lampname
                di[k]=v
        self._derived_properties["arclines"] = di
        

   


    



#! /usr/bin/env python
# -*- coding: utf-8 -*-

""" Internal small toolbox"""

import numpy as np

__all__ = ["kwargs_update","shape_ajustment", "is_arraylike"]


def kwargs_update(default,**kwargs):
    """
    """
    k = default.copy()
    for key,val in kwargs.items():
        k[key] = val
        
    return k

def load_pkl(filename):
    """
    """
    try:
        import cPickle as pkl
    except:
        import pickle
        with open(filename, 'rb') as f:
            u = pickle._Unpickler(f)
            u.encoding = 'latin1'
            return u.load()

    pkl_file = open(filename,'rb')
    return pkl.load(pkl_file)


def dump_pkl(data,filename,**kwargs):
    """
    """
    try:
        from cPickle import dump
    except:
        from pickle import dump
        
    if len(filename.split("."))>1 and filename.split(".")[-1]=="pkl":
        outfile =  open(filename,"wb")
    else:
        outfile =  open(filename+".pkl","wb")
    
    dump(data, outfile,**kwargs)
    outfile.close()

def is_arraylike(a):
    """ Tests if 'a' is an array / list / tuple """
    return isinstance(a, (list, tuple, np.ndarray) )


def fig_backend_test(backup='Agg'):
    """ """
    try:
        import matplotlib.pyplot as mpl
        fig = mpl.figure()
    except:
        import matplotlib
        matplotlib.use(backup)

def make_method(obj):
    """Decorator to make the function a method of *obj*.

    In the current context::
      @make_method(Axes)
      def toto(ax, ...):
          ...
    makes *toto* a method of `Axes`, so that one can directly use::
      ax.toto()
    COPYRIGHT: from Yannick Copin
    """

    def decorate(f):
        setattr(obj, f.__name__, f)
        return f

    return decorate

def shape_ajustment(x,y,model_x,**kwargs):
    """ creates a new version of `y` that matches the conversion x->model_x
    - This uses scipy.interpolate.UnivariateSpline -

    Parameters
    ----------
    x,y: [arrays]
        2 array of the same dimension.
        y is the array you want to reshape as x-> model_x

    model_x: [array]
        Model on which x will be projected.
        
    **kwargs goes to UnivariateSpline:
       - k (degree of smoothing <=5) etc.

    Returns
    -------
    array (same size as model_x)
    """
    from scipy.interpolate import UnivariateSpline

    #flagx = (x>=np.nanmin(model_x)) & (x<=np.nanmax(model_x))
    yrebin = UnivariateSpline(x, y,**kwargs)(model_x)
    
    if len(yrebin)==len(model_x):
        return yrebin
    
    warnings.warn("[shape_adjustment] shape does not match automatically. Removed edges")
    yrebinOK = N.empty((len(yrebin)+1),)
    yrebinOK[1:] = yrebin
    yrebinOK[0]  = yrebin[0]
    
    return yrebinOK

def is_running_from_notebook():
    """ Test if currently ran in notebook """
    return running_from() == "notebook"

def running_from():
    """  Where is the code running from ?"""
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return "notebook"   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return "terminal"  # Terminal running IPython
        else:
            return "other"  # Other type (?)
    except NameError:
        return None


    
def _loading_multiprocess():
    """ Black magic function to load to enable pickle in multiprocess """
    try:
        import copy_reg as copyreg
    except:
        import copyreg
    import types
    
    def _pickle_method(m):
        if m.im_self is None:
            return getattr, (m.im_class, m.im_func.func_name)
        else:
            return getattr, (m.im_self, m.im_func.func_name)
        
    copyreg.pickle(types.MethodType, _pickle_method)


def vac_to_air_sdss(vac):
    """ converts air wavelength [AA] into vacuum wavelengths. Tested. """
    return vac / (1.0 + 2.735182*10**-4 + 131.4182 / vac**2 + 2.76249*10**8 / vac**4)

def fetch_vac(x):
    from scipy.optimize import fmin
    def _to_be_min_(x_):
        return np.abs(x - tools.vac_to_air_sdss(x_))
    return fmin(_to_be_min_, x, disp=0)



def fit_intrinsic(data, model, errors, dof, intrinsic_guess=None):
    """ Get the most optimal intrinsic dispersion given the current fitted standardization parameters. 
    
    The optimal intrinsic magnitude dispersion is the value that has to be added in quadrature to 
    the errors such that the chi2/dof is 1.
    Returns
    -------
    float (intrinsic dispersion)
    """
    from scipy.optimize import fmin
    def get_intrinsic_chi2dof(intrinsic):
        return np.abs( np.nansum((data-model)**2/(errors**2+intrinsic**2)) / dof - 1)
    
    if intrinsic_guess is None:
        intrinsic_guess = np.nanmedian(errors)
        
    return fmin(get_intrinsic_chi2dof, intrinsic_guess, disp=0)[0]


################################
#                              #
#    MPL Like                  #
#                              #
################################
def draw_ellipse(image, xy, a, b, facecolor, edgecolor):
    """Improved ellipse drawing function, based on PIL.ImageDraw."""
    from PIL import Image, ImageDraw
    mask = Image.new( size=image.size, mode='RGBA')
    draw = ImageDraw.Draw(mask)
    draw.ellipse([xy[0]-a, xy[0]-b, xy[1]+a, xy[1]+b],
                     fill=tuple(facecolor), outline=tuple(edgecolor))
    mask = mask.resize(image.size, Image.LANCZOS)
    mask = mask.rotate(45)
    image.paste(mask, mask=mask)
    return image



""" Dasked version of pysedm/bin/extractstars.py """

import os
import numpy as np
from dask import delayed
from .base import DaskCCD


from .. import get_sedmcube, fluxcalibration, io


print("Dask Extract Cube to be implemented")
        
        
